#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from copy import deepcopy
from typing import Any, Optional, Tuple

from flsim.channels.base_channel import IdentityChannel
from flsim.clients.base_client import Client, ClientConfig
from flsim.common.timeout_simulator import TimeOutSimulator
from flsim.interfaces.model import IFLModel
from flsim.utils.async_trainer.async_user_selector import (
    AsyncUserSelector,
    AsyncUserSelectorInfo,
)
from flsim.utils.async_trainer.device_state import (
    DeviceState,
    TrainingSchedule,
    TrainingScheduleFactory,
    TrainingState,
)
from flsim.utils.async_trainer.training_event_generator import (
    IEventGenerator,
)
from flsim.utils.cuda import (
    ICudaStateManager,
    DEFAULT_CUDA_MANAGER,
)
from omegaconf import OmegaConf


class AsyncClientFactory:
    @classmethod
    def create(
        cls,
        current_time: float,
        event_generator: IEventGenerator,
        user_selector: AsyncUserSelector,
        client_config: ClientConfig,
        cuda_manager: ICudaStateManager = DEFAULT_CUDA_MANAGER,
        timeout_simulator: Optional[TimeOutSimulator] = None,
        channel: Optional[IdentityChannel] = None,
    ):
        user_info = user_selector.get_random_user()
        training_schedule = TrainingScheduleFactory.create(
            current_time, event_generator, user_info.user_data.num_examples()
        )
        client = Client(
            **OmegaConf.structured(client_config),
            dataset=user_info.user_data,
            name=f"client_{user_info.user_index}",
            timeout_simulator=timeout_simulator,
            channel=channel,
            cuda_manager=cuda_manager,
        )
        return AsyncClientDevice(training_schedule, client, user_info)


class AsyncClientDevice(DeviceState):
    r"""
    Class to represent a single async device. This class is responsible for
    maintaining the training state and training the local model
    """

    def __init__(
        self,
        training_schedule: TrainingSchedule,
        client: Client,
        user_info: AsyncUserSelectorInfo,
    ):
        self.client: Client = client
        self.local_model: IFLModel = None  # pyre-ignore[8]
        self.model_seqnum: int = -1
        self.user_info = user_info
        self.training_schedule = training_schedule
        super().__init__(training_schedule)

    def training_started(
        self, model_seqnum: int, init_model: Optional[IFLModel] = None
    ) -> None:
        r"""
        Starts the client training event by saving a copy of the current global model and seqnum
        """
        if init_model is not None:
            self.local_model = self.client.receive_through_channel(init_model)

        super().training_started()
        self.model_seqnum = model_seqnum

    def train_local_model(
        self, metric_reporter: Any = None
    ) -> Tuple[IFLModel, IFLModel, float]:
        r"""
        Performs local training loop
        """
        assert (
            self.local_model is not None
        ), "Client has not started training, local_model is None"
        # 1. Save the init model to compute delta
        before_train_local = deepcopy(self.local_model)
        # 2. Get ready for training
        self.local_model, optim, optim_scheduler = self.client.prepare_for_training(
            self.local_model
        )
        # 3. Train model on local data
        after_train_local, weight = self.client.train(
            self.local_model, optim, optim_scheduler, metric_reporter
        )
        # 3. Compute delta
        delta = self.client.compute_delta(
            before_train_local, after_train_local, model_to_save=before_train_local
        )
        # 4. Track client models if specified by config
        self.client.track(delta=delta, weight=weight, optimizer=optim)
        return delta, after_train_local, weight

    def is_waiting_to_start(self):
        return self.training_state == TrainingState.WAITING_FOR_START



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from flsim.active_user_selectors.simple_user_selector import (
    SequentialActiveUserSelectorConfig,
)
from flsim.channels.base_channel import FLChannelConfig
from flsim.clients.base_client import ClientConfig
from flsim.common.timeout_simulator import (
    NeverTimeOutSimulatorConfig,
    TimeOutSimulatorConfig,
)
from flsim.interfaces.model import IFLModel
from flsim.optimizers.local_optimizers import LocalOptimizerSGDConfig
from flsim.optimizers.optimizer_scheduler import (
    OptimizerSchedulerConfig,
    ConstantLRSchedulerConfig,
)
from flsim.servers.sync_servers import SyncServerConfig
from flsim.trainers.sync_trainer import SyncTrainer, SyncTrainerConfig
from omegaconf import OmegaConf

# have to create a variable because python linter doesn't like performing function calls
# in argument defaults (B008, https://github.com/PyCQA/flake8-bugbear#list-of-warnings)
NEVER_TIMEOUT_CONFIG = NeverTimeOutSimulatorConfig()
CONSTANT_LR_SCHEDULER_CONFIG = ConstantLRSchedulerConfig()
FED_AVG_SYNC_SERVER_CONFIG = SyncServerConfig(
    active_user_selector=SequentialActiveUserSelectorConfig()
)


def create_sync_trainer(
    model: IFLModel,
    local_lr: float,
    users_per_round: int,
    epochs: int,
    user_epochs_per_round: int = 1,
    do_eval: bool = True,
    server_config: SyncServerConfig = FED_AVG_SYNC_SERVER_CONFIG,
    timeout_simulator_config: TimeOutSimulatorConfig = NEVER_TIMEOUT_CONFIG,
    local_lr_scheduler: OptimizerSchedulerConfig = CONSTANT_LR_SCHEDULER_CONFIG,
    report_train_metrics: bool = False,
    dropout_rate: float = 1.0,
):
    # first disable report_train_metrics_after_aggregation. we will call
    # it outside of train() afterwise the post aggregation train metrics is
    # not returned
    sync_trainer = SyncTrainer(
        model=model,
        cuda_enabled=False,
        **OmegaConf.structured(
            SyncTrainerConfig(
                epochs=epochs,
                do_eval=do_eval,
                always_keep_trained_model=False,
                timeout_simulator=timeout_simulator_config,
                train_metrics_reported_per_epoch=1,
                eval_epoch_frequency=1,
                report_train_metrics=report_train_metrics,
                report_train_metrics_after_aggregation=False,
                client=ClientConfig(
                    epochs=user_epochs_per_round,
                    optimizer=LocalOptimizerSGDConfig(
                        lr=local_lr,
                    ),
                    lr_scheduler=local_lr_scheduler,
                    shuffle_batch_order=False,
                ),
                channel=FLChannelConfig(),
                server=server_config,
                users_per_round=users_per_round,
                dropout_rate=dropout_rate,
            )
        ),
    )
    return sync_trainer



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import abc
from dataclasses import dataclass

from omegaconf import MISSING
from torch.utils.data import Dataset


class FLDataset(Dataset, abc.ABC):
    """Abstract class that all Dataset for FL will be implementing for
    stronger type-checking and better abstraction. (e.g. FLDatasetDataLoader)
    """

    pass


@dataclass
class DatasetConfig:
    _target_: str = MISSING



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from typing import Type

import numpy as np
from flsim.utils.async_trainer.training_event_generator import (
    AsyncTrainingEventGenerator,
    AsyncTrainingEventGeneratorConfig,
    AsyncTrainingEventGeneratorFromList,
    AsyncTrainingEventGeneratorFromListConfig,
    AsyncTrainingStartTimeDistrConfig,
    ConstantAsyncTrainingStartTimeDistr,
    ConstantAsyncTrainingStartTimeDistrConfig,
    EventTimingInfo,
    IEventGenerator,
    PoissonAsyncTrainingStartTimeDistr,
    PoissonAsyncTrainingStartTimeDistrConfig,
)
from flsim.utils.timing.training_duration_distribution import (
    PerExampleGaussianDurationDistribution,
    PerExampleGaussianDurationDistributionConfig,
    PerUserGaussianDurationDistribution,
    PerUserGaussianDurationDistributionConfig,
)
from libfb.py import testutil
from omegaconf import OmegaConf


class EventDistributionsUtilTest(testutil.BaseFacebookTestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_simulated_training_training_event_generator(self) -> None:
        """Check that EventDistributionFromList works correctly by inputing
        a sample distribution, and confirming that the output is correct
        """
        timing_info1 = EventTimingInfo(prev_event_start_to_current_start=1, duration=2)
        timing_info2 = EventTimingInfo(prev_event_start_to_current_start=2, duration=1)
        timing_info3 = EventTimingInfo(prev_event_start_to_current_start=2, duration=5)
        random_list = [timing_info1, timing_info2, timing_info3]
        distr = AsyncTrainingEventGeneratorFromList(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorFromListConfig(training_events=random_list)
            )
        )
        self.assertTrue(
            distr.time_to_next_event_start()
            == timing_info1.prev_event_start_to_current_start
        )
        self.assertTrue(
            distr.training_duration(num_training_examples=1) == timing_info1.duration
        )
        self.assertTrue(
            distr.time_to_next_event_start()
            == timing_info2.prev_event_start_to_current_start
        )
        self.assertTrue(
            distr.training_duration(num_training_examples=1) == timing_info2.duration
        )
        self.assertTrue(
            distr.time_to_next_event_start()
            == timing_info3.prev_event_start_to_current_start
        )
        self.assertTrue(
            distr.training_duration(num_training_examples=1) == timing_info3.duration
        )

    def _duration_normality_check(
        self,
        event_generator: IEventGenerator,
        sample_count: int,
        expected_mean: float,
        expected_sd: float,
        epsilon: float,
    ) -> None:
        durations = []
        for _ in range(sample_count):
            durations.append(event_generator.training_duration(num_training_examples=1))
        # normality check doesn't verify mean and variance
        self.assertAlmostEqual(np.mean(durations), expected_mean, delta=epsilon)
        self.assertAlmostEqual(np.std(durations), expected_sd, delta=epsilon)

    def test_poisson_training_event_generator(self) -> None:
        """Check that TrainingEventDistritubion makes sense by checking that
        generated event durations follow normal distribution.
        """
        np.random.seed(1)
        # TODO: (jesikmin) T54593732 Add test to ensure that time_to_next_event_start
        # follows Poisson
        event_rate_per_sec, duration_mean, duration_sd = 10, 1, 5
        # set training_duration_min to a very negative number, to not bound the distribution
        duration_distr = PerExampleGaussianDurationDistributionConfig(
            training_duration_mean=duration_mean, training_duration_sd=duration_sd
        )
        training_start_time_distr = PoissonAsyncTrainingStartTimeDistrConfig(
            training_rate=event_rate_per_sec
        )
        distr = AsyncTrainingEventGenerator(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorConfig(
                    training_start_time_distribution=training_start_time_distr,
                    duration_distribution_generator=duration_distr,
                )
            )
        )
        self._duration_normality_check(
            distr,
            1000,  # sample_count
            duration_mean,
            duration_sd,
            epsilon=1,
        )

    def test_constant_training_event_distribution(self) -> None:
        """Check that ConstantAsyncTrainingStartTimeDistr generates the right
        next_event_time
        """
        np.random.seed(1)
        min_mean = 0.0001
        max_mean = 10
        max_sd = 1
        event_rate_per_sec = np.random.uniform(min_mean, max_mean)
        duration_mean = np.random.uniform(0, max_mean)
        duration_sd = np.random.uniform(0, max_sd)
        training_start_time_distr = ConstantAsyncTrainingStartTimeDistrConfig(
            training_rate=event_rate_per_sec
        )
        duration_distr = PerExampleGaussianDurationDistributionConfig(
            training_duration_mean=duration_mean, training_duration_sd=duration_sd
        )
        distr = AsyncTrainingEventGenerator(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorConfig(
                    training_start_time_distribution=training_start_time_distr,
                    duration_distribution_generator=duration_distr,
                )
            )
        )
        self.assertEqual(distr.time_to_next_event_start(), 1 / event_rate_per_sec)
        self._duration_normality_check(
            distr,
            1000,  # sample_count
            duration_mean,
            duration_sd,
            epsilon=1,
        )

    def test_constant_training_event_distribution_zero_sd(self) -> None:
        """Check that ConstantAsyncTrainingStartTimeDistr generates the right
        constant training_duration and next_event_time when SD is zero
        """
        np.random.seed(1)
        min_mean = 0.0001
        max_mean = 10
        for _num_rand_vals in range(10):
            event_rate_per_sec = np.random.uniform(min_mean, max_mean)
            duration_mean = np.random.uniform(0, max_mean)
            training_start_time_distr = ConstantAsyncTrainingStartTimeDistrConfig(
                training_rate=event_rate_per_sec
            )
            duration_distr = PerExampleGaussianDurationDistributionConfig(
                training_duration_mean=duration_mean, training_duration_sd=0
            )
            distr = AsyncTrainingEventGenerator(
                **OmegaConf.structured(
                    AsyncTrainingEventGeneratorConfig(
                        training_start_time_distribution=training_start_time_distr,
                        duration_distribution_generator=duration_distr,
                    )
                )
            )
            # generate duration and time_to_next_event_start two times
            for _ in range(2):
                self.assertAlmostEqual(
                    distr.training_duration(num_training_examples=1),
                    duration_mean,
                    delta=1e-4,
                )
                self.assertEqual(
                    distr.time_to_next_event_start(), 1 / event_rate_per_sec
                )

    def test_training_duration_min_bound(self) -> None:
        """Check that training_duration_min bound is followed"""
        np.random.seed(1)
        max_mean = 10
        max_sd = 1
        # generate 10 random event generators
        for _num_rand_vals in range(10):
            duration_mean = np.random.uniform(0, max_mean)
            duration_sd = np.random.uniform(0, max_sd)
            # choose a duration_min that is likely to be hit often
            duration_min = duration_mean
            per_user_duration_distr = PerUserGaussianDurationDistribution(
                **OmegaConf.structured(
                    PerUserGaussianDurationDistributionConfig(
                        training_duration_mean=duration_mean,
                        training_duration_sd=duration_sd,
                        training_duration_min=duration_min,
                    )
                )
            )
            per_example_duration_distr = PerExampleGaussianDurationDistribution(
                **OmegaConf.structured(
                    PerExampleGaussianDurationDistributionConfig(
                        training_duration_mean=duration_mean,
                        training_duration_sd=duration_sd,
                        training_duration_min=duration_min,
                    )
                )
            )
            # generate 100 random num_examples
            for _ in range(100):
                num_examples = np.random.randint(low=1, high=1e5)
                per_user_gaussian_duration = per_user_duration_distr.training_duration(
                    num_training_examples=num_examples
                )
                self.assertGreaterEqual(per_user_gaussian_duration, duration_min)
                # for per-example training duration, duration-min_bound applies to each example
                # while training_duration() returns time for each user
                # so actual bound is
                per_example_gaussian_duration = (
                    per_example_duration_distr.training_duration(
                        num_training_examples=num_examples
                    )
                )
                self.assertGreaterEqual(
                    per_example_gaussian_duration, duration_min * num_examples
                )

    @testutil.data_provider(
        lambda: (
            {
                "start_time_distr_config_class": PoissonAsyncTrainingStartTimeDistrConfig,
                "start_time_distr_class": PoissonAsyncTrainingStartTimeDistr,
            },
            {
                "start_time_distr_config_class": ConstantAsyncTrainingStartTimeDistrConfig,
                "start_time_distr_class": ConstantAsyncTrainingStartTimeDistr,
            },
        )
    )
    def test_string_conversion(
        self,
        start_time_distr_config_class: AsyncTrainingStartTimeDistrConfig,
        start_time_distr_class: Type,
    ) -> None:
        """Check that strings are correctly converted to TrainingEventGenerator"""
        training_rate = 1
        duration_mean_sec = 1
        training_duration_sd = 1
        training_start_time_distr = OmegaConf.structured(start_time_distr_config_class)
        training_start_time_distr.training_rate = training_rate
        duration_distr = PerExampleGaussianDurationDistributionConfig(
            training_duration_mean=duration_mean_sec,
            training_duration_sd=training_duration_sd,
        )
        self.assertEqual(
            AsyncTrainingEventGenerator(
                **OmegaConf.structured(
                    AsyncTrainingEventGeneratorConfig(
                        training_start_time_distribution=training_start_time_distr,
                        duration_distribution_generator=duration_distr,
                    )
                )
            )._training_start_time_distr.__class__,
            start_time_distr_class,
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

from dataclasses import dataclass
from itertools import chain
from typing import Optional, Tuple

from flsim.channels.base_channel import IdentityChannel
from flsim.interfaces.model import IFLModel
from flsim.privacy.common import PrivacyBudget, PrivacySetting
from flsim.privacy.privacy_engine import IPrivacyEngine
from flsim.privacy.privacy_engine_factory import PrivacyEngineFactory, NoiseType
from flsim.privacy.user_update_clip import UserUpdateClipper
from flsim.reducers.secure_round_reducer import (
    SecureRoundReducer,
    SecureRoundReducerConfig,
)
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.distributed.fl_distributed import FLDistributedUtils, OperationType
from flsim.utils.fl.common import FLModelParamUtils
from torch import nn


class DPRoundReducer(SecureRoundReducer):
    def __init__(
        self,
        *,
        global_model: IFLModel,
        num_users_per_round: int,
        total_number_of_users: int,
        channel: Optional[IdentityChannel] = None,
        name: Optional[str] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=DPRoundReducerConfig,
            **kwargs,
        )

        super().__init__(
            global_model=global_model,
            num_users_per_round=num_users_per_round,
            total_number_of_users=total_number_of_users,
            channel=channel,
            name=name,
            **kwargs,
        )
        self.num_users_per_round = num_users_per_round

        self.privacy_on = (
            # pyre-ignore[16]
            self.cfg.privacy_setting.noise_multiplier >= 0
            and self.cfg.privacy_setting.clipping_value < float("inf")
        )
        self.clipping_value = self.cfg.privacy_setting.clipping_value
        self.user_update_clipper = UserUpdateClipper(self.dtype)
        if self.privacy_on:
            self.privacy_engine: IPrivacyEngine = PrivacyEngineFactory.create(
                self.cfg.privacy_setting,
                num_users_per_round,
                total_number_of_users,
                noise_type=self.cfg.noise_type,
            )
            self.privacy_engine.attach(global_model=self.ref_model.fl_get_module())
        self._privacy_budget = PrivacyBudget(
            delta=self.cfg.privacy_setting.target_delta
        )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def update_reduced_module(self, delta_module: nn.Module, weight: float) -> None:
        """
        Please refer to ``RoundReducer.update_reduced_module`` for more info.
        Notes
        -----
        """
        if self.privacy_on:
            self.user_update_clipper.clip(delta_module, self.clipping_value)
        super().update_reduced_module(delta_module, weight)

    def reduce(self) -> Tuple[nn.Module, float]:
        if not self.privacy_on:
            return super().reduce()
        # only sum in the rank 0 reducer (no broadcast yet)
        self._reduce_all(OperationType.SUM)  # OperationType.SUM)
        self.logger.debug(f"Sum of weights after aggregation: {self.sum_weights}")
        if FLDistributedUtils.is_master_worker():
            total_weights = float(self.sum_weights.item())
            if abs(total_weights - self.num_users_per_round) > 1e-5:
                self.logger.error(
                    f"total weights {total_weights} is not equal to "
                    f"number of users {self.num_users_per_round}. "
                    "Please make sure reduction_type=AVERGAE."
                )

            """
            The final amount of noise added must be equal to
            (max_norm * noise_multiplier) / users_per_round, similar to
            Google's user-level DP https://arxiv.org/pdf/1710.06963.pdf.
            Note that in the _generate_noise() function, the noise_multiplier
            is already multiplied.
            """
            self.privacy_engine.add_noise(
                self.reduced_module, self.clipping_value / total_weights
            )

        # broadcast the new noisy model to all workers.
        state_dict = FLModelParamUtils.get_state_dict(
            self.reduced_module,
            # pyre-fixme[16]: `DPRoundReducer` has no attribute `cfg`.
            only_federated_params=self.cfg.only_federated_params,
        )
        FLDistributedUtils.distributed_operation(
            chain([self.sum_weights], state_dict.values()), OperationType.BROADCAST
        )
        self.logger.debug(
            f"Sum of client weights after reduction on worker: {self.sum_weights}"
        )
        self._privacy_budget = self.privacy_engine.get_privacy_spent()
        self.logger.info(f"User Privacy Budget: {self._privacy_budget}")
        return self.reduced_module, float(self.sum_weights.item())

    @property
    def privacy_budget(self) -> PrivacyBudget:
        return self._privacy_budget

    def reset(self, ref_model: IFLModel) -> None:
        super().reset(ref_model)
        if self.privacy_on:
            self.privacy_engine.attach(global_model=self.ref_model.fl_get_module())


@dataclass
class DPRoundReducerConfig(SecureRoundReducerConfig):
    """
    Contains configurations for a round reducer based on DP (user-level dp)
    """

    _target_: str = fullclassname(DPRoundReducer)
    privacy_setting: PrivacySetting = PrivacySetting()
    noise_type: NoiseType = NoiseType.GAUSSIAN



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
import math
from typing import List, Optional

import torch
from flsim.utils.fl.personalized_model import FLModelWithPrivateModules
from torch import nn
from torch.optim.optimizer import Optimizer


class FLModelParamUtils:
    @classmethod
    def get_state_dict(cls, model: nn.Module, only_federated_params: bool):
        if only_federated_params and isinstance(model, FLModelWithPrivateModules):
            state_dict = model.federated_state_dict()
        else:
            state_dict = model.state_dict()
        return state_dict

    @classmethod
    def load_state_dict(cls, model: nn.Module, state_dict, only_federated_params: bool):
        if only_federated_params and isinstance(model, FLModelWithPrivateModules):
            model.load_federated_state_dict(state_dict)
        else:
            model.load_state_dict(state_dict)

    @classmethod
    def zero_weights(cls, model: nn.Module, only_federated_params=False) -> None:
        state_dict = cls.get_state_dict(model, only_federated_params)
        for _name, param in state_dict.items():
            param.data.fill_(0.0)

    @classmethod
    def pseudo_random_weights(cls, model: nn.Module, seed: int = 1) -> None:
        torch.manual_seed(seed)
        for _name, param in model.state_dict().items():
            param.data.uniform_()

    @classmethod
    def get_mismatched_param(
        cls,
        models: List[nn.Module],
        rel_epsilon: Optional[float] = None,
        abs_epsilon: Optional[float] = None,
    ) -> str:
        """Compare all the models in the given list of models.
        It returns an empty string if all the models have same parameters.
        It returns the name of the first parameter that is different if any.
        """
        if rel_epsilon is None and abs_epsilon is not None:
            print("WARNING: rel_epsilon is not specified, abs_epsilon is ignored.")

        if len(models) <= 1:
            return ""
        dicts = [aModel.state_dict() for aModel in models]
        # verify new models have all params same
        rtol_atol = {}
        if rel_epsilon is not None:
            rtol_atol["rtol"] = rel_epsilon
        if abs_epsilon is not None:
            rtol_atol["atol"] = abs_epsilon
        for name, param in dicts[0].items():
            for adict in dicts[1:]:
                # if a parameter name does not exist in a model, return early
                if name not in adict.keys():
                    return name
                param_here = adict[name]
                # if epsilon is specified, do approx comparison
                if not torch.allclose(param.float(), param_here.float(), **rtol_atol):
                    return name
        return ""

    @classmethod
    def linear_comb_models(
        cls,
        model1: nn.Module,
        wt1: float,
        model2: nn.Module,
        wt2: float,
        model_to_save: nn.Module,
        only_federated_params: bool = False,
    ) -> None:
        """sets model_to_save = model1*wt1 + model2*wt2"""
        global_params = cls.get_state_dict(model_to_save, only_federated_params)
        params_model1 = cls.get_state_dict(model1, only_federated_params)
        params_model2 = cls.get_state_dict(model2, only_federated_params)

        assert (
            global_params.keys() == params_model1.keys() == params_model2.keys()
        ), "Models should have the same set of parameters, including order."

        with torch.no_grad():
            for name, global_param in global_params.items():
                global_param.data = (
                    params_model1[name].data * wt1 + params_model2[name].data * wt2
                )

        cls.load_state_dict(model_to_save, global_params, only_federated_params)

    @classmethod
    def average_models(
        cls,
        models: List[nn.Module],
        model_to_save: nn.Module,
        weights: Optional[List[float]] = None,
    ) -> None:
        """Averages parameters of input models. Saves the average model in model_to_save

        Args:
            models: collection of models. These will be changed in-place
            model_to_save: update this model with the average
            weights: (optional) use weighted average
        Returns:
            none
        """
        assert weights is None or len(weights) == len(models), (
            "Weights should have same length as models. len(wts):"
            + str(len(weights))
            + ", len(models):"
            + str(len(models))
        )
        wts_divisor = len(models)
        if weights is not None:
            for w in weights:
                assert w >= 0, "Weights must be non-negative. Found:" + str(w)
            wts_divisor = sum(weights)
            assert wts_divisor > 0, "Sum of weights must be positive:" + str(weights)
        cls.zero_weights(model_to_save, only_federated_params=True)
        for idx, aModel in enumerate(models):
            wts_numerator = 1 if weights is None else weights[idx]
            wt = wts_numerator / wts_divisor
            cls.linear_comb_models(
                aModel, wt, model_to_save, 1, model_to_save, only_federated_params=True
            )

    @classmethod
    def copy_models(
        cls,
        from_model: nn.Module,
        to_models: List[nn.Module],
        only_federated_params: bool = False,
    ) -> None:
        """Copy from_model into every model in to_models

        Args:
            from_model: a model
            to_models: collection of models. These will be changed in-place
            only_federated_params: copy only federated params.
        Returns:
            none
        """
        from_state_dict = cls.get_state_dict(from_model, only_federated_params)
        for m in to_models:
            cls.load_state_dict(m, from_state_dict, only_federated_params)

    @classmethod
    def clone(cls, model: nn.Module, dtype: Optional[torch.dtype] = None):
        """
        Clones a pytorch module, and allows for a change of precision.
        TODO If needed we can also add device here.
        """
        new_model = copy.deepcopy(model)
        return (
            new_model.float()  # pyre-ignore
            if dtype == torch.float32
            else (
                new_model.double()  # pyre-ignore
                if dtype == torch.float64
                else new_model
            )
        )

    @classmethod
    def set_gradient(cls, model: nn.Module, reference_gradient: nn.Module) -> None:
        """Set gradient of model to the parameters of reference_gradient
        Args:
            model: nn.Module
            reference_gradient: nn.Module - gradient is the parameters of this model
        """
        # Use parameters() since state_dict() may inlude non-learnable params.
        for m, ref in zip(model.parameters(), reference_gradient.parameters()):
            m.grad = ref.detach().clone().type(m.type())

    @classmethod
    def reconstruct_gradient(
        cls, old_model: nn.Module, new_model: nn.Module, grads: nn.Module
    ) -> None:
        # compute approximate gradient:
        # grads = old_model - new_model
        cls.subtract_model(old_model, new_model, grads)

    @classmethod
    def get_trainable_params(cls, model: nn.Module):
        return filter(lambda p: p.requires_grad, model.parameters())

    @classmethod
    def get_gradient_l2_norm_raw(cls, model: nn.Module) -> float:
        total_norm = 0
        for p in cls.get_trainable_params(model):
            if p.grad is None:
                continue
            param_norm = p.grad.data.norm(2)
            total_norm += param_norm.item() ** 2
        total_norm = total_norm ** (1.0 / 2)
        return total_norm

    @classmethod
    def get_num_trainable_params(cls, model: nn.Module) -> int:
        total_params = 0
        for p in cls.get_trainable_params(model):
            total_params += p.numel()
        return total_params

    @classmethod
    def get_gradient_l2_norm_normalized(cls, model: nn.Module) -> float:
        """Compute l2-norm-of-gradient/sqrt(num-params)
        If gradients are all independent, l2 norm grows as sqrt() of number
        of parameters. Eg: in Xavier Initialization
        """
        return cls.get_gradient_l2_norm_raw(model) / math.sqrt(
            cls.get_num_trainable_params(model)
        )

    @classmethod
    def debug_model_norm(cls, model: nn.Module):
        norm = 0
        for p in model.parameters():
            norm += torch.sum(torch.abs(p))
        return norm

    @classmethod
    def get_mismatched_param_max_difference(cls, models: List[nn.Module]):

        if len(models) <= 1:
            return 0.0
        dicts = [aModel.state_dict() for aModel in models]
        max_diff = 0
        # compute maximum element-wise difference of model parameters
        for name, param in dicts[0].items():
            for adict in dicts[1:]:
                param_here = adict[name]
                param_diff = torch.max(torch.abs(param - param_here)).item()
                # pyre-fixme[58]: `<` is not supported for operand types
                #  `Union[float, int]` and `int`.
                max_diff = param_diff if (param_diff > max_diff) else max_diff
                # if epsilon is specified, do approx comparison
        return max_diff

    @classmethod
    def clip_gradients(cls, max_normalized_l2_norm: float, model: nn.Module) -> None:
        """Clip gradients in model parameters by maximum value for normalized
        L2 norm (max_normalized_norm).
        """
        max_unnormalized_l2_norm = max_normalized_l2_norm * math.sqrt(
            cls.get_num_trainable_params(model)
        )
        torch.nn.utils.clip_grad_norm_(model.parameters(), max_unnormalized_l2_norm)

    @classmethod
    def step_with_modified_lr(
        cls, optimizer: Optimizer, base_lr: float, lr_normalizer: float
    ) -> None:
        for param_group in optimizer.param_groups:
            param_group["lr"] = base_lr * lr_normalizer
        # pyre-ignore[20] Call `Optimizer.step` expects argument `closure`
        # but closure is optional in torch.optim.Optimizer
        optimizer.step()

    @classmethod
    def multiply_model_by_weight(
        cls,
        model: nn.Module,
        weight: float,
        model_to_save: nn.Module,
        only_federated_params: bool = False,
    ):
        """
        Returns model_to_save = model * weight
        """
        FLModelParamUtils.linear_comb_models(
            model, weight, model, 0, model_to_save, only_federated_params
        )

    @classmethod
    def subtract_model(
        cls,
        minuend: nn.Module,
        subtrahend: nn.Module,
        difference: nn.Module,
        only_federated_params: bool = False,
    ):
        """
        Returns difference = minuend - subtrahend
        """
        FLModelParamUtils.linear_comb_models(
            minuend, 1, subtrahend, -1, difference, only_federated_params
        )

    @classmethod
    def add_model(
        cls,
        model1: nn.Module,
        model2: nn.Module,
        model_to_save: nn.Module,
        only_federated_params: bool = False,
    ):
        """
        Returns model_to_save = model1 + model2
        """
        FLModelParamUtils.linear_comb_models(
            model1, 1, model2, 1, model_to_save, only_federated_params
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import math
import string
from typing import Tuple

import torch
from flsim.data.data_provider import FLDataProviderFromList
from flsim.data.data_sharder import SequentialSharder, PowerLawSharder
from flsim.data.dataset_data_loader import FLDatasetDataLoaderWithBatch
from flsim.interfaces.model import IFLModel
from flsim.utils.sample_model import TestDataSetting
from torch.utils.data import Dataset


class Utils:
    @staticmethod
    def get_label(character: str) -> torch.Tensor:
        return torch.tensor(ord(character) % 2 == 0, dtype=torch.bool)

    @staticmethod
    def get_text(character: str) -> torch.Tensor:
        return torch.tensor(ord(character) - ord("a"), dtype=torch.long)

    @staticmethod
    def get_characters(num: int):
        """Return [a,b,c,d,.....z,a,b,c,d....] till we get num total chars"""
        characters = list(string.ascii_lowercase) * (1 + num // 26)
        return characters[:num]


class DummyAlphabetDataset(Dataset):
    """
    create a dummy PyTorch Dataset of k characters
    """

    def __init__(self, num_rows: int = 26):
        self.num_rows = num_rows
        self._data_rows = DummyAlphabetDataset.provide_data(self.num_rows)

    def __getitem__(self, index):
        return self._data_rows[index]

    def __len__(self):
        return self.num_rows

    @staticmethod
    def provide_data(num_rows: int = 26):
        """Generate num_row rows of data. Each row:
        "label":0 or 1, "text":int between 0 and 25
        """
        characters = Utils.get_characters(num_rows)
        return [
            {
                TestDataSetting.LABEL_COL_NAME: Utils.get_label(character),
                TestDataSetting.TEXT_COL_NAME: Utils.get_text(character),
            }
            for character in characters
        ]

    @staticmethod
    def create_data_provider_and_loader(
        dataset: Dataset, examples_per_user: int, batch_size: int, model
    ) -> Tuple[FLDataProviderFromList, FLDatasetDataLoaderWithBatch]:
        """
        Creates a data provider and data loader of type IFLDataProvider for a dataset
        """
        fl_data_sharder = SequentialSharder(examples_per_shard=examples_per_user)
        fl_data_loader = FLDatasetDataLoaderWithBatch(
            dataset,
            dataset,
            dataset,
            fl_data_sharder,
            batch_size,
            batch_size,
            batch_size,
        )
        fl_data_provider = FLDataProviderFromList(
            fl_data_loader.fl_train_set(),
            fl_data_loader.fl_eval_set(),
            fl_data_loader.fl_test_set(),
            model,
        )
        return fl_data_provider, fl_data_loader

    @staticmethod
    def create_data_provider_and_loader_uneven_split(
        num_examples: int,
        num_fl_users: int,
        batch_size: int,
        model: IFLModel,
        alpha: float = 0.2,
    ) -> Tuple[FLDataProviderFromList, FLDatasetDataLoaderWithBatch]:
        """
        Creates a data proivder and data loader with uneven number of
        examples per user following the power law distribution with order alpha
        """
        dataset = DummyAlphabetDataset(num_examples)
        fl_data_sharder = PowerLawSharder(num_shards=num_fl_users, alpha=alpha)
        fl_data_loader = FLDatasetDataLoaderWithBatch(
            dataset,
            dataset,
            dataset,
            fl_data_sharder,
            batch_size,
            batch_size,
            batch_size,
        )
        fl_data_provider = FLDataProviderFromList(
            fl_data_loader.fl_train_set(),
            fl_data_loader.fl_eval_set(),
            fl_data_loader.fl_test_set(),
            model,
        )
        return fl_data_provider, fl_data_loader


class NonOverlappingDataset(Dataset):
    """
    Create a dataset with non-overlapping non-zero entries for users. This will be useful
    for testing that clients will have orthogonal updates in linear regression problems.
    """

    def __init__(
        self,
        num_users: int = 10,
        num_nonzeros_per_user: int = 4,
        num_data_per_user: int = 6,
    ):
        self.num_users = num_users
        self.num_nonzeros_per_user = num_nonzeros_per_user
        self.num_data_per_user = num_data_per_user
        self._data_rows = NonOverlappingDataset.provide_data(
            num_users=self.num_users,
            num_nonzeros_per_user=self.num_nonzeros_per_user,
            num_data_per_user=self.num_data_per_user,
        )

    def __getitem__(self, index):
        return self._data_rows[index]

    def __len__(self):
        return self.num_data_per_user * self.num_users

    @staticmethod
    def provide_data(
        num_users: int = 10, num_nonzeros_per_user: int = 4, num_data_per_user: int = 6
    ):
        """
        Generate data. Each successive group of num_user_data rows has
        num_user_nonzeros non-overlapping non-zero entries
        """

        num_rows = num_data_per_user * num_users
        num_cols = num_nonzeros_per_user * num_users
        non_overlap_data = torch.zeros(num_rows, num_cols)

        for row in range(num_rows):
            col_start = math.floor(row / num_data_per_user) * num_nonzeros_per_user
            non_overlap_data[
                row, col_start : col_start + num_nonzeros_per_user
            ] = torch.rand(1, num_nonzeros_per_user)

        labels = torch.rand(num_rows, 1)

        return [
            {
                TestDataSetting.LABEL_COL_NAME: labels[row, :],
                TestDataSetting.TEXT_COL_NAME: non_overlap_data[row, :],
            }
            for row in range(num_rows)
        ]

    @staticmethod
    def create_data_provider_and_loader(
        dataset: Dataset, examples_per_user: int, batch_size: int, model: IFLModel
    ) -> Tuple[FLDataProviderFromList, FLDatasetDataLoaderWithBatch]:
        """
        Creates a data provider and data loader of type IFLDataProvider for a dataset
        """
        fl_data_sharder = SequentialSharder(examples_per_shard=examples_per_user)
        fl_data_loader = FLDatasetDataLoaderWithBatch(
            dataset,
            dataset,
            dataset,
            fl_data_sharder,
            batch_size,
            batch_size,
            batch_size,
        )
        fl_data_provider = FLDataProviderFromList(
            fl_data_loader.fl_train_set(),
            fl_data_loader.fl_eval_set(),
            fl_data_loader.fl_test_set(),
            model,
        )
        return fl_data_provider, fl_data_loader


class RandomDataset(Dataset):
    """
    Create a dataset with random entries and labels.
    """

    def __init__(
        self, num_users: int = 10, num_data_per_user: int = 6, dim_data: int = 40
    ):
        self.num_users = num_users
        self.num_data_per_user = num_data_per_user
        self.dim_data = dim_data
        self._data_rows = RandomDataset.provide_data(
            num_users=self.num_users,
            num_data_per_user=self.num_data_per_user,
            dim_data=self.dim_data,
        )

    def __getitem__(self, index):
        return self._data_rows[index]

    def __len__(self):
        return self.num_data_per_user * self.num_users

    @staticmethod
    def provide_data(
        num_users: int = 10, num_data_per_user: int = 6, dim_data: int = 40
    ):
        """
        Generate data which is a random matrix.
        """

        num_rows = num_data_per_user * num_users
        num_cols = dim_data
        random_data = torch.randn(num_rows, num_cols)
        labels = torch.rand(num_rows, 1)

        return [
            {
                TestDataSetting.LABEL_COL_NAME: labels[row, :],
                TestDataSetting.TEXT_COL_NAME: random_data[row, :],
            }
            for row in range(num_rows)
        ]

    @staticmethod
    def create_data_provider_and_loader(
        dataset: Dataset, examples_per_user: int, batch_size: int, model: IFLModel
    ) -> Tuple[FLDataProviderFromList, FLDatasetDataLoaderWithBatch]:
        """
        Creates a data provider and data loader of type IFLDataProvider for a dataset
        """
        fl_data_sharder = SequentialSharder(examples_per_shard=examples_per_user)
        fl_data_loader = FLDatasetDataLoaderWithBatch(
            dataset,
            dataset,
            dataset,
            fl_data_sharder,
            batch_size,
            batch_size,
            batch_size,
        )
        fl_data_provider = FLDataProviderFromList(
            fl_data_loader.fl_train_set(),
            fl_data_loader.fl_eval_set(),
            fl_data_loader.fl_test_set(),
            model,
        )
        return fl_data_provider, fl_data_loader



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import pytest
from flsim.common.pytest_helper import assertEqual, assertTrue
from flsim.optimizers.async_aggregators import (
    AsyncAggregatorConfig,
    FedAdamAsyncAggregatorConfig,
    FedAvgWithLRAsyncAggregatorConfig,
    HybridAggregatorConfig,
    FedAdamHybridAggregatorConfig,
    FedAvgWithLRHybridAggregatorConfig,
)
from flsim.optimizers.local_optimizers import LocalOptimizerSGDConfig
from flsim.optimizers.optimizer_test_utils import OptimizerTestUtil
from flsim.optimizers.sync_aggregators import (
    SyncAggregatorConfig,
    FedAdamSyncAggregatorConfig,
    FedAvgWithLRSyncAggregatorConfig,
    FedAvgSyncAggregatorConfig,
)
from flsim.utils.sample_model import TwoLayerNet
from hydra.utils import instantiate
from omegaconf import OmegaConf


@pytest.fixture(scope="class")
def prepare_optimizer_factory_test(request):
    request.cls.model = TwoLayerNet(10, 5, 1)


@pytest.mark.usefixtures("prepare_optimizer_factory_test")
class TestOptimizerFactory:
    @pytest.mark.parametrize(
        "type_str,config", OptimizerTestUtil.provide_sync_factory_creation_dataset()
    )
    def test_sync_optimizer_config_creation_through_hydra(
        self, type_str: str, config: SyncAggregatorConfig
    ) -> None:
        if type_str == "FedAvg":
            assertTrue(
                # pyre-fixme[16]: Optional type has no attribute `__name__`.
                OmegaConf.get_type(config).__name__,
                FedAvgSyncAggregatorConfig.__name__,
            )
        if type_str == "FedAvgWithLR":
            assertTrue(
                OmegaConf.get_type(config).__name__,
                FedAvgWithLRSyncAggregatorConfig.__name__,
            )
            # pyre-fixme[16]: `SyncAggregatorConfig` has no attribute `lr`.
            assertEqual(config.lr, 0.1)
        if type_str == "FedAdam":
            assertTrue(
                OmegaConf.get_type(config).__name__,
                FedAdamSyncAggregatorConfig.__name__,
            )
            assertEqual(config.lr, 0.1)
            # pyre-fixme[16]: `SyncAggregatorConfig` has no attribute `weight_decay`.
            assertEqual(config.weight_decay, 0.9)
            # pyre-fixme[16]: `SyncAggregatorConfig` has no attribute `eps`.
            assertEqual(config.eps, 1e-8)

    @pytest.mark.parametrize(
        "type_str,config", OptimizerTestUtil.provide_async_factory_creation_dataset()
    )
    def test_async_optimizer_config_creation_through_hydra(
        self, type_str: str, config: AsyncAggregatorConfig
    ) -> None:
        if type_str == "FedAvgWithLR":
            assertTrue(
                # pyre-fixme[16]: Optional type has no attribute `__name__`.
                OmegaConf.get_type(config).__name__,
                FedAvgWithLRAsyncAggregatorConfig.__name__,
            )
            # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `lr`.
            assertEqual(config.lr, 0.1)
        if type_str == "FedAdam":
            assertTrue(
                OmegaConf.get_type(config).__name__,
                FedAdamAsyncAggregatorConfig.__name__,
            )
            assertEqual(config.lr, 0.1)
            # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `weight_decay`.
            assertEqual(config.weight_decay, 0.9)
            # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `eps`.
            assertEqual(config.eps, 1e-8)

    @pytest.mark.parametrize(
        "type_str,config", OptimizerTestUtil.provide_hybrid_factory_creation_dataset()
    )
    def test_hybrid_optimizer_config_creation_through_hydra(
        self, type_str: str, config: HybridAggregatorConfig
    ) -> None:
        if type_str == "FedAvgWithLR":
            assertTrue(
                # pyre-fixme[16]: Optional type has no attribute `__name__`.
                OmegaConf.get_type(config).__name__,
                FedAvgWithLRHybridAggregatorConfig.__name__,
            )
            # pyre-fixme[16]: `HybridAggregatorConfig` has no attribute `lr`.
            assertEqual(config.lr, 0.1)
            assertEqual(config.buffer_size, 3)
        if type_str == "FedAdam":
            assertTrue(
                OmegaConf.get_type(config).__name__,
                FedAdamHybridAggregatorConfig.__name__,
            )
            assertEqual(config.lr, 0.1)
            # pyre-fixme[16]: `HybridAggregatorConfig` has no attribute `weight_decay`.
            assertEqual(config.weight_decay, 0.9)
            # pyre-fixme[16]: `HybridAggregatorConfig` has no attribute `eps`.
            assertEqual(config.eps, 1e-8)
            assertEqual(config.buffer_size, 3)

    def test_local_optimizer_creation(self) -> None:
        config = {
            "_target_": LocalOptimizerSGDConfig._target_,
            "lr": 1.0,
            "weight_decay": 0.1,
        }
        # pyre-ignore[16]: for pytest fixture
        local_optimizer = instantiate(config, model=self.model)
        assertEqual(
            OptimizerTestUtil.get_value_from_optimizer(local_optimizer, "lr"),
            config["lr"],
        )
        assertEqual(
            OptimizerTestUtil.get_value_from_optimizer(local_optimizer, "weight_decay"),
            config["weight_decay"],
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import math
from collections import OrderedDict
from dataclasses import dataclass

from flsim.channels.base_channel import IdentityChannel, FLChannelConfig
from flsim.channels.message import Message
from flsim.compression.pq import PQ
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg


class ProductQuantizationChannel(IdentityChannel):
    """
    Implements a channel that emulates Product Quantization.
    The idea is to split the weight matrices (linear or
    convolutional) to a set of subvectors and to learn a
    codebook on these subvectors using k-means. More details
    on the procedure in the files em.py and pq.py. See paper for
    more details: https://arxiv.org/abs/1907.05686.

    Notes:
        - We do not quantize the biases since their compression
          overhead is very small.
        - We do not quantize small layers having less than
          `min_numel_to_quantize` elements.
        - There is the possibility to learn multiple codebooks
          per matrix by setting num_codebooks.
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=ProductQuantizationChannelConfig,
            **kwargs,
        )
        super().__init__(**kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _calc_message_size_client_to_server(self, message: Message):
        """
        We compute the size of the compressed message as follows:
            - for the weights (compressed):
                * log(n_centroids) / 8 bytes per element (for the assignment)
                * num_codebooks * block_size * n_centroids fp32 elements for the centroids
            - for the biases (not compressed): 4 bytes per element

        Notes:
            - n_centroids is not necessarily equal to max_num_centroids, hence we
              recover it from the shape of `param["centroids"]`.
        """
        message_size_bytes = 0
        for param in message.model_state_dict.values():
            # param was compressed with PQ
            if type(param) is dict:
                block_size = param["centroids"].size(1)
                n_subvectors = param["assignments"].size(0)
                n_centroids = param["centroids"].size(0) // self.cfg.num_codebooks

                assignments_bytes = math.log2(n_centroids) / 8.0 * n_subvectors
                centroids_bytes = (
                    self.cfg.num_codebooks
                    * n_centroids
                    * block_size
                    * ProductQuantizationChannel.BYTES_PER_FP32
                )
                message_size_bytes += assignments_bytes + centroids_bytes
            # param is a non-compressed torch.Tensor
            else:
                message_size_bytes += (
                    ProductQuantizationChannel.BYTES_PER_FP32 * param.numel()
                )
        return message_size_bytes

    def _on_client_before_transmission(self, message: Message) -> Message:
        """
        We quantize the weights under the form of centroids
        and assignments and do not quantize the biases.
        """
        message.populate_state_dict()
        new_state_dict = OrderedDict()
        for name, param in message.model_state_dict.items():
            # compress only large weight matrices
            # pyre-fixme[16]: `ProductQuantizationChannel` has no attribute `cfg`.
            if param.ndim > 1 and param.numel() >= self.cfg.min_numel_to_quantize:
                pq = PQ(
                    param.data.size(),
                    self.cfg.max_block_size,
                    self.cfg.num_codebooks,
                    self.cfg.max_num_centroids,
                    self.cfg.num_k_means_iter,
                    self.cfg.verbose,
                )
                centroids, assignments = pq.encode(param.data.cpu())
                compressed_param = {
                    "sizes": pq.sizes,
                    "centroids": centroids.data,
                    "assignments": assignments.data,
                }
                new_state_dict[name] = compressed_param
            # do not compress biases and small layers
            else:
                new_state_dict[name] = param.data

        message.model_state_dict = new_state_dict
        return message

    def _on_server_after_reception(self, message: Message) -> Message:
        """
        We reconstruct the weights from the centroids
        and the assignments.
        """

        new_state_dict = OrderedDict()
        for name, param in message.model_state_dict.items():
            # param was compressed with PQ. TODO: more robust check than `type(param)`
            if type(param) is dict:
                pq = PQ(
                    param["sizes"],
                    self.cfg.max_block_size,
                    self.cfg.num_codebooks,
                    self.cfg.max_num_centroids,
                    self.cfg.num_k_means_iter,
                    self.cfg.verbose,
                )
                decompressed_param = pq.decode(
                    param["centroids"].data, param["assignments"].data
                )
                new_state_dict[name] = decompressed_param
            # param is a non-compressed torch.Tensor
            else:
                new_state_dict[name] = param.data
        message.model_state_dict = new_state_dict
        message.update_model_()
        return message


@dataclass
class ProductQuantizationChannelConfig(FLChannelConfig):
    _target_: str = fullclassname(ProductQuantizationChannel)
    max_num_centroids: int = 256
    min_numel_to_quantize: int = 10
    num_codebooks: int = 1
    max_block_size: int = 9
    num_k_means_iter: int = 20
    verbose: bool = False



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import abc
from dataclasses import dataclass
from enum import Enum, auto

import numpy as np
from flsim.data.data_provider import IFLDataProvider, IFLUserData


@dataclass
class AsyncUserSelectorInfo:
    r"""
    Dataclass to encapsulate a selected user for async training

    user_data (IFLUserData): seleected user data in the dataset
    user_index (int): the index for user_data assuming IFLDataProvider.train_data is a List
    """
    user_data: IFLUserData
    user_index: int


class AsyncUserSelector(abc.ABC):
    def __init__(self, data_provider: IFLDataProvider):
        self.data_provider: IFLDataProvider = data_provider

    @abc.abstractmethod
    def get_random_user(self) -> AsyncUserSelectorInfo:
        r"""
        Returns a random IFLUserData from the dataset and the user index (for testing)
        """
        pass


class RandomAsyncUserSelector(AsyncUserSelector):
    def __init__(self, data_provider: IFLDataProvider):
        super().__init__(data_provider)

    def get_random_user(self) -> AsyncUserSelectorInfo:
        user_index = np.random.randint(0, self.data_provider.num_users())
        return AsyncUserSelectorInfo(self.data_provider[user_index], user_index)


class RoundRobinAsyncUserSelector(AsyncUserSelector):
    r"""
    Chooses users in round-robin order, starting from user=0.
    Particularly useful for testing.
    """

    def __init__(self, data_provider: IFLDataProvider):
        super().__init__(data_provider)
        self.current_user_index: int = 0

    def get_random_user(self) -> AsyncUserSelectorInfo:
        user_index = self.current_user_index
        self.current_user_index = (
            self.current_user_index + 1
        ) % self.data_provider.num_users()
        return AsyncUserSelectorInfo(self.data_provider[user_index], user_index)


class AsyncUserSelectorType(Enum):
    RANDOM = auto()
    ROUND_ROBIN = auto()


class AsyncUserSelectorFactory:
    @classmethod
    def create_users_selector(
        cls, type: AsyncUserSelectorType, data_provider: IFLDataProvider
    ):
        if type == AsyncUserSelectorType.RANDOM:
            return RandomAsyncUserSelector(data_provider)
        elif type == AsyncUserSelectorType.ROUND_ROBIN:
            return RoundRobinAsyncUserSelector(data_provider)
        else:
            raise AssertionError(f"Unknown user selector type: {type}")



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import abc
import logging
import sys
from dataclasses import dataclass
from typing import Any, Iterable, List, Optional, Tuple

import torch
from flsim.channels.base_channel import FLChannelConfig
from flsim.channels.communication_stats import (
    ChannelDirection,
)
from flsim.clients.base_client import ClientConfig
from flsim.common.logger import Logger
from flsim.common.timeline import Timeline
from flsim.common.timeout_simulator import (
    TimeOutSimulatorConfig,
    NeverTimeOutSimulatorConfig,
)
from flsim.data.data_provider import IFLDataProvider
from flsim.interfaces.metrics_reporter import IFLMetricsReporter, Metric, TrainingStage
from flsim.interfaces.model import IFLModel
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.cuda import CudaTransferMinimizer
from hydra.utils import instantiate
from omegaconf import MISSING
from omegaconf import OmegaConf


class FLTrainer(abc.ABC):
    """Base class for FederatedLearning Training"""

    logger: logging.Logger = Logger.get_logger(__name__)

    def __init__(
        self,
        *,
        model: IFLModel,
        cuda_enabled: bool = False,
        **kwargs,
    ):

        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=FLTrainerConfig,
            **kwargs,
        )

        assert (
            # pyre-fixme[16]: `FLTrainer` has no attribute `cfg`.
            self.cfg.eval_epoch_frequency
            <= self.cfg.epochs
        ), "We expect to do at least one eval. However, eval_epoch_frequency:"
        f" {self.cfg.eval_epoch_frequency} > total epochs: {self.cfg.epochs}"

        self._cuda_state_manager = CudaTransferMinimizer(cuda_enabled)
        self._cuda_state_manager.on_trainer_init(model)
        self.cuda_enabled = cuda_enabled

        self._timeout_simulator = instantiate(self.cfg.timeout_simulator)
        self.channel = instantiate(self.cfg.channel)
        self.data_provider = None
        self.num_total_users: int = -1
        # Initialize tracker for measuring communication between the clients and
        # the server if communication metrics are enabled

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        if OmegaConf.is_missing(cfg.timeout_simulator, "_target_"):
            cfg.timeout_simulator = NeverTimeOutSimulatorConfig()
        if OmegaConf.is_missing(cfg.client, "_target_"):
            cfg.client = ClientConfig()
        if OmegaConf.is_missing(cfg.channel, "_target_"):
            cfg.channel = FLChannelConfig()

    @abc.abstractmethod
    def train(
        self,
        data_provider: IFLDataProvider,
        metric_reporter: IFLMetricsReporter,
        num_total_users: int,
        distributed_world_size: int = 1,
        rank: int = 0,
    ) -> Tuple[IFLModel, Any]:
        pass

    def test(
        self, data_iter: Iterable[Any], metric_reporter: IFLMetricsReporter
    ) -> Any:
        return self._test(
            timeline=Timeline(global_round=1),
            data_iter=data_iter,
            model=self.global_model(),
            metric_reporter=metric_reporter,
        )

    def global_model(self) -> IFLModel:
        pass

    def _maybe_run_evaluation(
        self,
        model: IFLModel,
        timeline: Timeline,
        eval_iter: Iterable[Any],
        metric_reporter: IFLMetricsReporter,
        best_metric,
        best_model_state,
    ):
        # pyre-fixme[16]: `FLTrainer` has no attribute `cfg`.
        if not self.cfg.do_eval:
            return best_metric, best_model_state
        if not timeline.tick(self.cfg.eval_epoch_frequency):
            return best_metric, best_model_state
        eval_metric, eval_metric_better_than_prev = self._evaluate(
            timeline=timeline,
            data_iter=eval_iter,
            model=model,
            metric_reporter=metric_reporter,
        )

        # 1) keep the best model so far if metric_reporter.compare_metrics is specified
        # 2) if self.always_keep_trained_model is set as true, ignore the metricsa and
        #    keep the trained model for each epoch
        if self.cfg.always_keep_trained_model or eval_metric_better_than_prev:
            # last_best_epoch = epoch
            self.logger.info(
                f"Found a better model!, current_eval_metric:{eval_metric}"
            )
            best_metric = eval_metric
            model_state = model.fl_get_module().state_dict()
            best_model_state = model_state
        sys.stdout.flush()
        return best_metric, best_model_state

    def _print_training_stats(self, timeline: Timeline) -> None:
        print(f"Train finished Global Round: {timeline.global_round_num()}")

    def _report_train_metrics(
        self,
        model: IFLModel,
        timeline: Timeline,
        metric_reporter: Optional[IFLMetricsReporter] = None,
        extra_metrics: Optional[List[Metric]] = None,
    ) -> None:
        if (
            # pyre-fixme[16]: `FLTrainer` has no attribute `cfg`.
            self.cfg.report_train_metrics
            and metric_reporter is not None
            and timeline.tick(1.0 / self.cfg.train_metrics_reported_per_epoch)
        ):
            self._print_training_stats(timeline)
            metric_reporter.report_metrics(
                model=model,
                reset=True,
                stage=TrainingStage.TRAINING,
                timeline=timeline,
                epoch=timeline.global_round_num(),  # for legacy
                print_to_channels=True,
                extra_metrics=extra_metrics,
            )

    def _calc_post_epoch_communication_metrics(
        self, timeline: Timeline, metric_reporter: Optional[IFLMetricsReporter]
    ):

        if (
            metric_reporter is not None
            and self.channel.cfg.report_communication_metrics
            # pyre-fixme[16]: `FLTrainer` has no attribute `cfg`
            and timeline.tick(1.0 / self.cfg.train_metrics_reported_per_epoch)
        ):
            extra_metrics = [
                Metric(
                    "Client to Server Bytes Sent"
                    if name == ChannelDirection.CLIENT_TO_SERVER
                    else "Server to Client Bytes Sent",
                    tracker.mean(),
                )
                for name, tracker in self.channel.stats_collector.get_channel_stats().items()
            ]
            metric_reporter.report_metrics(
                model=None,
                reset=False,
                stage=TrainingStage.TRAINING,
                timeline=timeline,
                epoch=timeline.global_round_num(),  # for legacy
                print_to_channels=True,
                extra_metrics=extra_metrics,
            )
            self.channel.stats_collector.reset_channel_stats()

    def _evaluate(
        self,
        timeline: Timeline,
        data_iter: Iterable[Any],
        model: IFLModel,
        metric_reporter: IFLMetricsReporter,
    ) -> Tuple[Any, bool]:
        with torch.no_grad():
            self._cuda_state_manager.before_train_or_eval(model)
            model.fl_get_module().eval()
            print(f"Running {timeline} for {TrainingStage.EVAL.name.title()}")

            for _, batch in enumerate(data_iter):
                batch_metrics = model.get_eval_metrics(batch)
                metric_reporter.add_batch_metrics(batch_metrics)

            metrics, found_best_model = metric_reporter.report_metrics(
                model=model,
                reset=True,
                stage=TrainingStage.EVAL,
                timeline=timeline,
                epoch=timeline.global_round_num(),  # for legacy
                print_to_channels=True,
            )
            self._cuda_state_manager.after_train_or_eval(model)
            return metrics, found_best_model

    def _test(
        self,
        timeline: Timeline,
        data_iter: Iterable[Any],
        model: IFLModel,
        metric_reporter: IFLMetricsReporter,
    ) -> Any:
        with torch.no_grad():
            self._cuda_state_manager.before_train_or_eval(model)
            model.fl_get_module().eval()
            print(f"Running {timeline} for {TrainingStage.TEST.name.title()}")

            for _, batch in enumerate(data_iter):
                batch_metrics = model.get_eval_metrics(batch)
                metric_reporter.add_batch_metrics(batch_metrics)

            metrics, _ = metric_reporter.report_metrics(
                model=model,
                reset=True,
                stage=TrainingStage.TEST,
                timeline=timeline,
                epoch=timeline.global_round_num(),  # for legacy
                print_to_channels=True,
            )
            self._cuda_state_manager.after_train_or_eval(model)
            return metrics


@dataclass
class FLTrainerConfig:
    # TODO: (jesikmin) T59233747 [papaya][trainer] Remove default values and
    # fix all tests
    _target_: str = MISSING
    _recursive_: bool = False
    # Training epochs
    epochs: float = 10.0
    # Whether to do evaluation and model selection based on it.
    do_eval: bool = True
    # don't use metric reporter to choose: always keep trained model
    always_keep_trained_model: bool = False
    # client training timeout
    timeout_simulator: TimeOutSimulatorConfig = TimeOutSimulatorConfig()
    # how many times per epoch should we report training metrics
    # numbers greater than 1 help with plotting more precise training curves
    train_metrics_reported_per_epoch: int = 1
    # perform eval to do model selection in every eval_epoch_frequency epochs
    eval_epoch_frequency: float = 1.0
    # Whether metrics on training data should be computed and reported.
    report_train_metrics: bool = True
    report_train_metrics_after_aggregation: bool = False
    use_train_clients_for_aggregation_metrics: bool = True
    # config for the clients
    client: ClientConfig = ClientConfig()
    # config for the channels
    channel: FLChannelConfig = FLChannelConfig()



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
import math
import unittest
from unittest.mock import MagicMock

import numpy as np
import torch
import torch.nn as nn
from flsim.common.pytest_helper import (
    assertEqual,
    assertNotEqual,
    assertAlmostEqual,
    assertRaises,
    assertFalse,
    assertLessEqual,
    assertTrue,
)
from flsim.privacy.common import PrivacySetting
from flsim.privacy.privacy_engine import (
    GaussianPrivacyEngine,
    TreePrivacyEngine,
    PrivacyEngineNotAttachedException,
)
from flsim.privacy.privacy_engine_factory import PrivacyEngineFactory, NoiseType
from flsim.tests import utils
from libfb.py import testutil
from opacus import privacy_analysis


class TestGaussianPrivacyEngine:
    def _init_privacy_engine(
        self,
        alphas=[1 + x / 10.0 for x in range(1, 100)],
        noise_multiplier=1.0,
        target_delta=1e-5,
        users_per_round=10,
        num_total_users=10,
        global_model_parameter_val=5.0,
        noise_seed=0,
    ):
        privacy_setting = PrivacySetting(
            alphas=alphas,
            noise_multiplier=noise_multiplier,
            target_delta=target_delta,
            noise_seed=noise_seed,
        )
        privacy_engine = GaussianPrivacyEngine(
            privacy_setting=privacy_setting,
            users_per_round=users_per_round,
            num_total_users=num_total_users,
        )
        global_model = utils.TwoFC()  # This model has 15 weights and 6 biases
        global_model.fill_all(global_model_parameter_val)
        privacy_engine.attach(global_model)
        return privacy_engine

    def _calc_eps(self, sample_rate, noise_multiplier, steps, alphas, delta):
        rdp = privacy_analysis.compute_rdp(sample_rate, noise_multiplier, steps, alphas)
        eps, _ = privacy_analysis.get_privacy_spent(alphas, rdp, delta=delta)
        return eps

    def test_privacy_analysis_alpha_in_alphas(self):
        """
        Tests that the optimal alphas of DP analysis is in the range
        of the original alphas.
        """
        privacy_engine = self._init_privacy_engine()
        privacy_budget = privacy_engine.get_privacy_spent()
        assertTrue(privacy_budget.alpha in privacy_engine.alphas)

    def test_privacy_analysis_epsilon_reasonable(self):
        """
        Tests that the epsilon is greater than 1 in normal settings.
        Also, when we do not add any noise, the privacy loss should
        be infinity.
        """

        privacy_engine = self._init_privacy_engine()
        privacy_budget = privacy_engine.get_privacy_spent()
        assertTrue(privacy_budget.epsilon > 0)

        privacy_engine.noise_multiplier = 0
        privacy_budget = privacy_engine.get_privacy_spent()
        assertTrue(privacy_budget.epsilon == float("inf"))

    def test_privacy_analysis_epsilon(self):
        """
        Tests that the epsilon calculations are correct
        """

        alphas = [1 + x / 10.0 for x in range(1, 100)]
        noise_multiplier = 1.5
        target_delta = 1e-5

        num_users = 1000
        num_users_per_round = 50
        steps = num_users // num_users_per_round
        user_sampling_rate = num_users_per_round / num_users

        privacy_engine = self._init_privacy_engine(
            alphas=alphas,
            noise_multiplier=noise_multiplier,
            target_delta=target_delta,
            num_total_users=num_users,
            users_per_round=num_users_per_round,
        )
        model_diff = utils.TwoFC()
        for _ in range(steps):  # adding noise will increse the steps
            privacy_engine.add_noise(model_diff, 1.0)

        privacy_budget = privacy_engine.get_privacy_spent()
        eps = self._calc_eps(
            user_sampling_rate, noise_multiplier, steps, alphas, target_delta
        )
        assertEqual(privacy_budget.epsilon, eps)

    def test_noise_added(self):
        """
        Tests that noise is successfully added to a model update, by
        checking that the model update after noise addition is different
        from the original model update.
        """
        model_diff = utils.TwoFC()  # model update
        model_diff.fill_all(1.0)

        model_diff_before_noise = copy.deepcopy(model_diff)

        privacy_engine = self._init_privacy_engine()
        privacy_engine.add_noise(model_diff, sensitivity=0.5)

        mismatched = utils.verify_models_equivalent_after_training(
            model_diff_before_noise, model_diff
        )
        assertNotEqual(mismatched, "")

    def test_deterministic_noise_addition(self):
        """
        Tests when the noise seed is set to the same value, we get
        the same (i.e. deterministic) noise-added model updates. It
        also tests when the seed is set to different values, we will
        get different noise-added model updates.
        """
        model_diff = utils.TwoFC()  # model update
        model_diff.fill_all(1.0)

        model_diff_another_seed = copy.deepcopy(model_diff)
        model_diff_same_seed = copy.deepcopy(model_diff)

        privacy_engine = self._init_privacy_engine(noise_seed=1003)
        privacy_engine.add_noise(model_diff, sensitivity=0.5)

        privacy_engine = self._init_privacy_engine(noise_seed=2000)
        privacy_engine.add_noise(model_diff_another_seed, sensitivity=0.5)

        mismatched = utils.verify_models_equivalent_after_training(
            model_diff, model_diff_another_seed
        )
        assertNotEqual(mismatched, "")

        privacy_engine = self._init_privacy_engine(noise_seed=1003)
        privacy_engine.add_noise(model_diff_same_seed, sensitivity=0.5)

        mismatched = utils.verify_models_equivalent_after_training(
            model_diff, model_diff_same_seed
        )
        assertEqual(mismatched, "")

    def test_not_attached_validator(self):
        """
        Tests that the Privacy Engine throws a not attach
        exception if it is not properly attached.
        """
        model_diff = utils.TwoFC()  # model update
        model_diff.fill_all(1.0)

        privacy_setting = PrivacySetting(
            alphas=[1 + x / 10.0 for x in range(1, 100)],
            noise_multiplier=1.0,
            target_delta=1e-6,
        )
        privacy_engine = GaussianPrivacyEngine(
            privacy_setting=privacy_setting, users_per_round=1, num_total_users=1
        )
        sensitivity = 0.5
        with assertRaises(PrivacyEngineNotAttachedException):
            privacy_engine.add_noise(model_diff, sensitivity)

        raised_exception = False
        global_model = utils.TwoFC()
        global_model.fill_all(5.0)
        privacy_engine.attach(global_model)
        try:
            privacy_engine.add_noise(model_diff, sensitivity)
        except PrivacyEngineNotAttachedException:
            raised_exception = True
        assertFalse(raised_exception)


class TreePrivacyEngineTest(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

    def _create_delta(self, dim, value=0.0):
        delta = nn.Linear(dim, 1)
        delta.bias.data.fill_(value)
        delta.weight.data.fill_(value)
        return delta, copy.deepcopy(delta)

    def _count_bits(self, n: int):
        """
        Returns the number of
        1s in the binary representations of i
        """
        count = 0
        while n:
            n &= n - 1
            count += 1
        return count

    @testutil.data_provider(
        lambda: (
            {"num_leaf": 4, "max_height": 2},
            {"num_leaf": 8, "max_height": 3},
            {"num_leaf": 16, "max_height": 4},
        )
    )
    def test_build_tree(self, num_leaf, max_height):
        """
        Test that build tree logic is correct.
        For any binary tree with n leaves, the tree's height
        will be log2(n) tall
        """
        tree = TreePrivacyEngine.build_tree(num_leaf)
        for node in tree:
            assertLessEqual(node.height, math.ceil(math.log2(num_leaf)))

    def test_basic_tree_node_sensitivity(self):
        """
        This is a small test. If we set the noise in each node as 1,
        we should be seeing the returned noise as the number of
        1s in the binary representations of ith step
        """

        def generate_noise(*args):
            return 1

        num_steps = 32
        tree = TreePrivacyEngine(
            PrivacySetting(noise_multiplier=1.0, noise_seed=0),
            users_per_round=1,
            num_total_users=num_steps,
            efficient_tree=False,
        )
        tree._generate_noise = MagicMock(side_effect=generate_noise)
        for i in range(num_steps):
            cumsum = tree.range_sum(0, i, size=torch.Size([1]), sensitivity=1.0)
            bits = float(self._count_bits(i + 1))
            assertEqual(cumsum, bits)

    @testutil.data_provider(
        lambda: (
            {"upr": 4, "n_users": 4, "noise_multiplier": 1, "exp_var": 0.57},
            {"upr": 7, "n_users": 7, "noise_multiplier": 1, "exp_var": 2.23},
            {"upr": 8, "n_users": 8, "noise_multiplier": 1, "exp_var": 0.53},
            {"upr": 8, "n_users": 8, "noise_multiplier": 2, "exp_var": 2.13},
            {"upr": 8, "n_users": 8, "noise_multiplier": 0.5, "exp_var": 0.13},
        )
    )
    def test_tree_noise_sum_expected(self, upr, n_users, noise_multiplier, exp_var):
        """
        See D27796316 test plan for explanation
        """
        delta, _ = self._create_delta(dim=1000, value=0)

        setting = PrivacySetting(
            noise_multiplier=noise_multiplier,
            noise_seed=0,
        )
        privacy_engine = PrivacyEngineFactory.create(
            setting,
            users_per_round=upr,
            num_total_users=n_users,
            noise_type=NoiseType.TREE_NOISE,
        )

        privacy_engine.add_noise(delta, sensitivity=1.0)
        noised_delta = torch.flatten(
            torch.stack([p for name, p in delta.named_parameters() if "weight" in name])
        )
        assertAlmostEqual(torch.var(noised_delta), exp_var, delta=0.15)

    @testutil.data_provider(
        lambda: (
            {"steps": 4, "upr": 4, "n_users": 4, "sigma": 1, "exp_var": 0.57},
            {"steps": 7, "upr": 7, "n_users": 7, "sigma": 1, "exp_var": 2.23},
            {"steps": 8, "upr": 8, "n_users": 8, "sigma": 1, "exp_var": 0.53},
            {"steps": 8, "upr": 8, "n_users": 8, "sigma": 2, "exp_var": 2.13},
            {"steps": 8, "upr": 8, "n_users": 8, "sigma": 0.5, "exp_var": 0.13},
        )
    )
    def test_range_sum_noise_expected(self, steps, upr, n_users, sigma, exp_var):
        """
        See D27796316 test plan for explanation
        """
        setting = PrivacySetting(
            noise_multiplier=sigma,
            noise_seed=1,
        )
        privacy_engine = PrivacyEngineFactory.create(
            setting,
            users_per_round=upr,
            num_total_users=n_users,
            noise_type=NoiseType.TREE_NOISE,
        )
        for i in range(steps):
            sum_ = privacy_engine.range_sum(0, i, torch.Size((1000,)), 1.0)

        assertAlmostEqual(torch.var(sum_), exp_var, delta=0.15)

    @testutil.data_provider(
        lambda: (
            {"n_users": 1600, "upr": 100, "sigma": 4.03, "epsilon": 4.19},
            {"n_users": 1600, "upr": 100, "sigma": 6.21, "epsilon": 2.60},
            {"n_users": 1600, "upr": 100, "sigma": 8.83, "epsilon": 1.77},
        )
    )
    def test_privacy_analysis_epsilon(self, n_users, upr, sigma, epsilon):
        delta, _ = self._create_delta(dim=1000, value=0)
        setting = PrivacySetting(
            noise_multiplier=sigma,
            noise_seed=1,
            alphas=np.arange(1.01, 100, 0.01).tolist(),
            target_delta=1e-6,
        )
        privacy_engine = PrivacyEngineFactory.create(
            setting,
            users_per_round=upr,
            num_total_users=n_users,
            noise_type=NoiseType.TREE_NOISE,
        )
        for _ in range(n_users // upr):
            privacy_engine.add_noise(delta, sensitivity=1.0)
        budget = privacy_engine.get_privacy_spent()
        assertAlmostEqual(budget.epsilon, epsilon, delta=0.5)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.


from tempfile import mkstemp

import pytest
import torch.distributed as dist
import torch.multiprocessing as mp
from flsim.common.pytest_helper import (
    assertEqual,
    assertNotEqual,
    assertAlmostEqual,
    assertEmpty,
)
from flsim.servers.aggregator import Aggregator, AggregationType
from flsim.tests.utils import (
    create_model_with_value,
    model_parameters_equal_to_value,
)
from flsim.utils.distributed.fl_distributed import FLDistributedUtils, OperationType


def init_process(
    rank,
    world_size,
    aggregator,
    models,
    file_loc,
    pipe,
    distributed_op,
):
    FLDistributedUtils.dist_init(
        rank=rank,
        world_size=world_size,
        init_method=f"file://{file_loc}",
        use_cuda=False,
    )
    aggregator.zero_weights()
    for i, m in enumerate(models):
        if i % world_size == rank:
            weight = i + 1
            aggregator.apply_weight_to_update(delta=m, weight=weight)
            aggregator.add_update(m, weight=weight)
    module = aggregator.aggregate(distributed_op)
    sums, weights = 0.0, 0.0
    all_sum = [(p.sum(), p.numel()) for p in module.parameters()]
    for s, w in all_sum:
        sums += float(s)
        weights += float(w)
    pipe.send(sums / weights)
    dist.destroy_process_group()


def run_multiprocess_aggregation_test(
    aggregator,
    num_processes=1,
    num_models=4,
    distributed_op=OperationType.SUM_AND_BROADCAST,
):
    _, tmpfile = mkstemp(dir="/tmp")
    pipe_out, pipe_in = mp.Pipe(False)
    models = [create_model_with_value(1.0) for i in range(num_models)]

    processes = []
    results = []
    FLDistributedUtils.WORLD_SIZE = num_processes
    for pid in range(num_processes):
        p = mp.Process(
            target=init_process,
            args=(
                pid,
                num_processes,
                aggregator,
                models,
                tmpfile,
                pipe_in,
                distributed_op,
            ),
        )
        p.start()
        processes.append(p)
        results.append(pipe_out)
    for p in processes:
        p.join()
    res = [r.recv() for r in results]
    return res


AGGREGATION_TYPES = [
    AggregationType.AVERAGE,
    AggregationType.WEIGHTED_AVERAGE,
    AggregationType.SUM,
    AggregationType.WEIGHTED_SUM,
]


class TestAggregator:
    def test_zero_weights(self) -> None:
        model = create_model_with_value(0)
        ag = Aggregator(module=model, aggregation_type=AggregationType.AVERAGE)
        weight = 1.0
        steps = 5

        for _ in range(steps):
            delta = create_model_with_value(1.0)
            ag.apply_weight_to_update(delta=delta, weight=weight)
            ag.add_update(delta=delta, weight=weight)
        assertEqual(ag.sum_weights.item(), weight * steps)

        ag.zero_weights()
        assertEqual(ag.sum_weights.item(), 0)

    @pytest.mark.parametrize(
        "agg_type,num_process,num_models,expected_value",
        [
            (AggregationType.AVERAGE, 4, 10, 1.0),
            (AggregationType.WEIGHTED_AVERAGE, 4, 10, 1.0),
            (AggregationType.WEIGHTED_SUM, 4, 10, 55.0),
            (AggregationType.SUM, 4, 10, 10.0),
        ],
    )
    def test_multiprocess_aggregation(
        self, agg_type, num_process, num_models, expected_value
    ):
        model = create_model_with_value(0)
        ag = Aggregator(module=model, aggregation_type=agg_type)

        results = run_multiprocess_aggregation_test(
            ag, num_processes=num_process, num_models=num_models
        )
        for result in results:
            assertAlmostEqual(result, expected_value, places=5)

    @pytest.mark.parametrize(
        "agg_type,expected_value",
        [
            (AggregationType.AVERAGE, 1.0),
            (AggregationType.WEIGHTED_AVERAGE, 1.0),
            (AggregationType.WEIGHTED_SUM, 55.0),
            (AggregationType.SUM, 10.0),
        ],
    )
    def test_aggregate(self, agg_type, expected_value):

        model = create_model_with_value(0)
        ag = Aggregator(module=model, aggregation_type=agg_type)

        ag.zero_weights()
        for i in range(10):
            delta = create_model_with_value(1.0)
            weight = i + 1
            ag.apply_weight_to_update(delta=delta, weight=weight)
            ag.add_update(delta=delta, weight=weight)

        model = ag.aggregate()
        error_msg = model_parameters_equal_to_value(model, expected_value)
        assertEmpty(error_msg, msg=error_msg)

    @pytest.mark.parametrize(
        "agg_type,expected_value",
        [
            (AggregationType.AVERAGE, 10.0),
            (AggregationType.WEIGHTED_AVERAGE, 55.0),
            (AggregationType.WEIGHTED_SUM, 55.0),
            (AggregationType.SUM, 10.0),
        ],
    )
    def test_add_update(self, agg_type, expected_value):
        model = create_model_with_value(0)
        ag = Aggregator(module=model, aggregation_type=agg_type)
        ag.zero_weights()
        for i in range(10):
            delta = create_model_with_value(1.0)
            weight = i + 1
            ag.apply_weight_to_update(delta=delta, weight=weight)
            ag.add_update(delta=delta, weight=weight)

        assertEqual(ag.sum_weights.item(), expected_value)

    @pytest.mark.parametrize(
        "agg_type,dist_op",
        [
            (AggregationType.AVERAGE, OperationType.SUM),
            (AggregationType.WEIGHTED_AVERAGE, OperationType.SUM),
            (AggregationType.WEIGHTED_SUM, OperationType.SUM),
            (AggregationType.SUM, OperationType.SUM),
        ],
    )
    def test_distributed_op_aggregation(self, agg_type, dist_op):
        """
        Test aggregation with only SUM and no BROADTCAST then each worker should have
        different parameters.
        """
        model = create_model_with_value(0)
        ag = Aggregator(module=model, aggregation_type=agg_type)

        results = run_multiprocess_aggregation_test(
            ag,
            num_processes=4,
            num_models=10,
            distributed_op=dist_op,
        )
        for r, v in zip(results, results[1:]):
            assertNotEqual(r, v)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import hashlib
import time
from collections import defaultdict
from itertools import zip_longest
from typing import Any, Dict, Generator, Iterable, List, Optional

import torch


def batchify(
    iterable: Iterable[Any], batch_size: int, drop_last: Optional[bool] = False
) -> Generator:
    """
    Groups list into batches
    Example:
    >>> batchify([1, 2, 3, 4, 5], 2)
    >>> [[1, 2], [3, 4], [5]]
    """
    iterators = [iter(iterable)] * batch_size
    for batch in zip_longest(*iterators, fillvalue=None):
        batch = [ex for ex in batch if ex is not None]
        if drop_last and len(batch) != batch_size:
            break
        yield batch


def merge_dicts(batch: List[Dict[str, torch.Tensor]]) -> Dict[str, torch.Tensor]:
    """
    Merge a list of dictionaries into one dictionary
    Example:
    >>> merge_dicts([{"a": torch.Tensor([1])}, {"a": torch.Tensor([2])}])
    >>> {"a": torch.Tensor([[1.0], [2.0]])},
    """
    res = defaultdict(list)
    for ex in batch:
        for key, value in ex.items():
            res[key].append(value)
    return {k: torch.stack(v) for k, v in res.items()}


def stable_hash(base: int = 100000) -> int:
    md5 = hashlib.md5(str(time.time()).encode("utf-8"))
    return int.from_bytes(md5.digest(), byteorder="little") % base



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from math import log10, sqrt

from flsim.utils.async_trainer.async_example_weights import (
    EqualExampleWeightConfig,
    EqualExampleWeight,
    ExampleWeight,
    LinearExampleWeightConfig,
    LinearExampleWeight,
    Log10ExampleWeightConfig,
    Log10ExampleWeight,
    SqrtExampleWeightConfig,
    SqrtExampleWeight,
)
from flsim.utils.async_trainer.async_staleness_weights import (
    ConstantStalenessWeightConfig,
    ConstantStalenessWeight,
    PolynomialStalenessWeightConfig,
    PolynomialStalenessWeight,
    ThresholdStalenessWeightConfig,
    ThresholdStalenessWeight,
)


class AsyncExampleWeightsTestUtils:
    EXAMPLE_WEIGHT_TEST_CONFIGS = [
        (EqualExampleWeightConfig(), EqualExampleWeight),
        (LinearExampleWeightConfig(), LinearExampleWeight),
        (SqrtExampleWeightConfig(), SqrtExampleWeight),
        (Log10ExampleWeightConfig(), Log10ExampleWeight),
    ]

    AVG_NUMBER_OF_EXAMPLES = [1, 10000]

    @classmethod
    def expected_weight(
        cls,
        avg_num_examples: int,
        num_examples: int,
        example_weight_class: ExampleWeight,
    ) -> float:
        if example_weight_class == EqualExampleWeight:
            return 1.0
        elif example_weight_class == LinearExampleWeight:
            return num_examples / avg_num_examples
        elif example_weight_class == SqrtExampleWeight:
            return sqrt(num_examples) / sqrt(avg_num_examples)
        elif example_weight_class == Log10ExampleWeight:
            return log10(1 + num_examples) / log10(1 + avg_num_examples)
        else:
            raise AssertionError(f"Unknown example_weight type:{example_weight_class}")


class AsyncStalenessWeightsTestUtils:
    STALENESS_WEIGHT_TEST_CONFIGS = [
        (ConstantStalenessWeightConfig(), ConstantStalenessWeight),
        (
            ThresholdStalenessWeightConfig(cutoff=1, value_after_cutoff=0.1),
            ThresholdStalenessWeight,
        ),
        (PolynomialStalenessWeightConfig(exponent=0.5), PolynomialStalenessWeight),
    ]
    AVG_TEST_STALENESS = [1, 10000]

    @classmethod
    def get_constant_wt(cls) -> float:
        return 1.0

    @classmethod
    def get_threshold_wt(
        cls, staleness: int, cutoff: int, value_after_cutoff: float
    ) -> float:
        return 1.0 if staleness <= cutoff else value_after_cutoff

    @classmethod
    def get_polynomial_wt(cls, staleness: int, exponent: float) -> float:
        return 1 / ((1 + staleness) ** exponent)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import logging
import math
from dataclasses import dataclass
from typing import Any, List, Tuple, Optional

from flsim.clients.async_client import AsyncClientDevice
from flsim.common.logger import Logger
from flsim.common.timeline import Timeline
from flsim.common.training_event_handler import AsyncTrainingEventHandler
from flsim.common.training_simulator import AsyncTrainingSimulator
from flsim.data.data_provider import IFLDataProvider
from flsim.interfaces.metrics_reporter import IFLMetricsReporter, Metric
from flsim.interfaces.model import IFLModel
from flsim.optimizers.async_aggregators import (
    AsyncAggregatorConfig,
    FedAvgWithLRAsyncAggregatorConfig,
    AsyncAggregator,
    HybridAggregator,
)
from flsim.trainers.trainer_base import FLTrainer, FLTrainerConfig
from flsim.utils.async_trainer.async_user_selector import (
    AsyncUserSelectorFactory,
    AsyncUserSelectorType,
)
from flsim.utils.async_trainer.async_weights import AsyncWeight, AsyncWeightConfig
from flsim.utils.async_trainer.training_event_generator import (
    AsyncTrainingEventGeneratorConfig,
    EventGeneratorConfig,
)
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.cuda import GPUMemoryMinimizer, CudaTransferMinimizer
from hydra.utils import instantiate
from omegaconf import OmegaConf
from tqdm import tqdm


class AsyncTrainer(FLTrainer, AsyncTrainingEventHandler):
    """Implements Async version of FederatedAveraging
    Implemented like AsynchronousSGD, except that we don't have real gradients,
    but we use reconstructed gradients as if they're real gradients

    Attributes:
        epochs (int): Training epochs
        report_train_metrics (bool): Whether metrics on training data should be
            computed and reported.
    Internal attributes:
    """

    logger: logging.Logger = Logger.get_logger(__name__)

    def __init__(
        self,
        *,
        model: IFLModel,
        cuda_enabled: bool = False,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=AsyncTrainerConfig,
            **kwargs,
        )

        super().__init__(model=model, cuda_enabled=cuda_enabled, **kwargs)
        self._cuda_state_manager = (
            CudaTransferMinimizer(cuda_enabled)
            # pyre-fixme[16]: `AsyncTrainer` has no attribute `cfg`.
            if self.cfg.minimize_cuda_transfer
            else GPUMemoryMinimizer(cuda_enabled)
        )
        self._cuda_state_manager.on_trainer_init(model)
        self.aggregator: AsyncAggregator = instantiate(
            self.cfg.aggregator, global_model=model, channel=self.channel
        )
        AsyncTrainingEventHandler.__init__(self)
        self._training_simulator: Optional[AsyncTrainingSimulator] = None
        self.weight: AsyncWeight = instantiate(self.cfg.async_weight)
        self._event_generator = instantiate(
            self.cfg.training_event_generator,
        )
        # for pyre; declare instance variables (https://fburl.com/88n6i71r)
        # pyre-fixme[8]: Attribute has type `IFLMetricsReporter`; used as `None`.
        self.metric_reporter = None  # type: IFLMetricsReporter
        self.best_metric = None
        self.best_model_state = self.global_model().fl_get_module().state_dict()

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        if OmegaConf.is_missing(cfg.aggregator, "_target_"):
            cfg.aggregator = FedAvgWithLRAsyncAggregatorConfig()
        if OmegaConf.is_missing(cfg.training_event_generator, "_target_"):
            cfg.training_event_generator = AsyncTrainingEventGeneratorConfig()
        if OmegaConf.is_missing(cfg.async_weight, "_target_"):
            cfg.async_weight = AsyncWeightConfig()

    def global_model(self) -> IFLModel:
        """This function makes it explicit that self.global_model() is owned
        by the aggregator, not by ASyncTrainer
        """
        return self.aggregator.global_model

    @property
    def global_round(self):
        return self.aggregator.global_seqnum + 1  # seqnum is 0 based.

    def on_training_start(self, client: AsyncClientDevice) -> None:
        r"""
        Overrides `AsyncTrainingEventHandler.on_training_start` to mark a client starting to train

        Marks the client state as training and copies the current global model
        to the client's local model

        Note:
        -----
        The actual training process doesn't start until on_training_end
        """
        client.training_started(
            model_seqnum=self.aggregator.global_seqnum, init_model=self.global_model()
        )

    def train_and_update_global_model(self, client: AsyncClientDevice) -> None:
        r"""
        Train a single client and aggregate update into the global model
        """
        assert client.local_model is not None
        # 1. zero grad global optimizer
        self.aggregator.zero_grad()
        # 2. train client on local data
        client_delta, final_local_model, num_examples = client.train_local_model(
            self.metric_reporter
        )
        assert num_examples > 0, "Client must have more than one example"
        # 3. get client staleness
        client_staleness = self.aggregator.model_staleness(
            model_seqnum=client.model_seqnum
        )
        # 4. check participation criteria
        if self._can_participate(staleness=client_staleness):
            weight = self.weight.weight(
                num_examples=num_examples, staleness=client_staleness
            )
            # 5. maybe take a global step
            is_global_model_updated = self.aggregator.on_client_training_end(
                client_delta=client_delta,
                final_local_model=final_local_model,
                weight=weight,
            )
            if is_global_model_updated:
                self._global_update_done()

    def _can_participate(self, staleness: float):
        # pyre-fixme[16]: `AsyncTrainer` has no attribute `cfg`.
        return staleness <= self.cfg.max_staleness

    def _num_global_steps_in_epoch(self):
        if isinstance(self.aggregator, HybridAggregator):
            return math.ceil(self.num_total_users / self.aggregator.cfg.buffer_size)
        else:
            return self.num_total_users

    def _global_update_done(self) -> None:
        timeline = self._get_timeline()
        self._report_train_metrics(
            model=self.global_model(),
            timeline=timeline,
            metric_reporter=self.metric_reporter,
            extra_metrics=self._get_training_metrics(),
        )
        self._calc_post_epoch_communication_metrics(
            self._get_timeline(), self.metric_reporter
        )
        (self.best_metric, self.best_model_state,) = FLTrainer._maybe_run_evaluation(
            self,
            self.global_model(),
            timeline,
            self.data_provider.eval_data(),
            self.metric_reporter,
            self.best_metric,
            self.best_model_state,
        )

    def _get_training_metrics(self) -> List[Metric]:
        metrics = Metric.from_args(
            # pyre-fixme[16]: `Optional` has no attribute `queue_stats`.
            Concurrency_Rate=self._training_simulator.queue_stats.avg_pending_jobs(),
            Seqnum_Diff_Mean=self.aggregator.seqnum_tracker.mean(),
            Seqnum_Diff_Std=self.aggregator.seqnum_tracker.standard_deviation(),
            Weight_Mean=self.weight.stats.mean(),
            Weight_Std=self.weight.stats.standard_deviation(),
        )
        return (
            # pyre-ignore[16]: if aggregator is private then we have privacy budget
            metrics + Metric.from_dict(self.aggregator.privacy_budget._asdict())
            if self.aggregator.is_private
            else metrics
        )

    def _get_timeline(self):
        return Timeline(
            global_round=self.global_round,
            rounds_per_epoch=self._num_global_steps_in_epoch(),
        )

    def _print_training_stats(self, timeline: Timeline):
        super()._print_training_stats(timeline)
        print("Async trainer stats:")
        # pyre-fixme[16]: `Optional` has no attribute `print_stats`.
        self._training_simulator.print_stats("\t")
        self.aggregator.seqnum_tracker.print_stats()

    def train(
        self,
        data_provider: IFLDataProvider,
        metric_reporter: IFLMetricsReporter,
        num_total_users: int,
        distributed_world_size: int = 1,
        rank: int = 0,
    ) -> Tuple[IFLModel, Any]:
        """Train and eval a model, the model states will be modified.
        Args:
            train_iter (List[Iterable[Any]]): list of batch iterators of training data,
                each batch iterator represents data from a single 'user'
            eval_iter (Iterable[Any]): batch iterator of evaluation data
            model (Model): model to be trained
            metric_reporter (IFLMetricsReporter): compute metric based on training
                output and report results to console, file.. etc

        Returns:
            model, best_metric: the trained model together with the best metric
        """
        assert (
            rank == 0 and distributed_world_size == 1
        ), "Distributed training not supported yet for AsyncTrainer"

        self.best_metric = None
        self.best_model_state = self.global_model().fl_get_module().state_dict()
        self.data_provider = data_provider
        self.metric_reporter = metric_reporter
        self.num_total_users = data_provider.num_users()
        self.aggregator.set_num_total_users(self.num_total_users)
        user_selector = AsyncUserSelectorFactory.create_users_selector(
            # pyre-fixme[16]: `AsyncTrainer` has no attribute `cfg`.
            self.cfg.async_user_selector_type,
            data_provider,
        )

        self._training_simulator = AsyncTrainingSimulator(
            event_generator=self._event_generator,
            job_scheduler=self,
            user_selector=user_selector,
            num_train_end_events_per_epoch=self.num_total_users,
            shared_client_config=self.cfg.client,
            timeout_simulator=self._timeout_simulator,
            channel=self.channel,
            cuda_manager=self._cuda_state_manager,
        )
        num_int_epochs = math.ceil(self.cfg.epochs)
        for _epoch in tqdm(range(1, num_int_epochs + 1), desc="Epoch", unit="epoch"):
            # pyre-fixme[16]: `Optional` has no attribute `run_one_epoch`.
            self._training_simulator.run_one_epoch()
            # in k-async, up to (k-1) client updates may not have been aggregated
            # into the global model
            had_unaggregated_updates = self.aggregator.on_training_epoch_end()

            if had_unaggregated_updates:
                self._global_update_done()
        self.global_model().fl_get_module().load_state_dict(self.best_model_state)
        return self.global_model(), self.best_metric


@dataclass
class AsyncTrainerConfig(FLTrainerConfig):
    _target_: str = fullclassname(AsyncTrainer)
    aggregator: AsyncAggregatorConfig = AsyncAggregatorConfig()
    training_event_generator: EventGeneratorConfig = EventGeneratorConfig()
    # TODO: async_user_selector_type should be directly instantiable from json_config
    async_user_selector_type: AsyncUserSelectorType = AsyncUserSelectorType.RANDOM
    async_weight: AsyncWeightConfig = AsyncWeightConfig()
    # any client with staleness greater than this number will be rejected
    max_staleness: float = float("inf")
    # Minimize CPU<-->GPU memory bandwidth at the cost of increasing GPU memory consumption,
    # Turning this off will increase training time
    minimize_cuda_transfer: bool = True



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from flsim.common.pytest_helper import (
    assertEqual,
    assertAlmostEqual,
    assertFalse,
    assertTrue,
)
from flsim.common.timeline import Timeline


class TestTimeline:
    def test_global_round_num(self):
        # test first round
        tl = Timeline(epoch=1, round=1)
        assertEqual(tl.global_round_num(), 1)

        # test first round does not depned on rounds_per_epoch
        tl = Timeline(epoch=1, round=1, rounds_per_epoch=100)
        assertEqual(tl.global_round_num(), 1)

        # test global_round_num properly calculated if global_round not given
        tl = Timeline(epoch=2, round=3, rounds_per_epoch=4)
        assertEqual(tl.global_round_num(), 7)

        # test global_round_num equals global_round regardless of epoch/round
        tl = Timeline(epoch=1, round=1, rounds_per_epoch=1, global_round=123)
        assertEqual(tl.global_round_num(), tl.global_round)

    def test_as_float(self):
        # test first round
        tl = Timeline(epoch=1, round=1)
        assertAlmostEqual(tl.as_float(), 1.0)  # float starts from 1

        # test first round
        tl = Timeline(epoch=1, round=1, rounds_per_epoch=100)
        assertAlmostEqual(tl.as_float(), 1 / 100)

        # first round of 10th epoch
        tl = Timeline(epoch=10, round=1)
        assertAlmostEqual(tl.as_float(), 10.0)

        # test first round does not depned on rounds_per_epoch
        tl = Timeline(epoch=10, round=1, rounds_per_epoch=100)
        assertAlmostEqual(tl.as_float(), 9 + (1 / 100))

        # test offset
        tl = Timeline(epoch=2, round=1, rounds_per_epoch=100)
        assertAlmostEqual(tl.as_float(-1), 1.0)

        # test from global_round
        tl = Timeline(global_round=12, rounds_per_epoch=10)
        assertAlmostEqual(tl.as_float(), 1.2)
        assertAlmostEqual(tl.as_float(1), 1.3)

        # test global_round dominates
        tl = Timeline(global_round=12, rounds_per_epoch=10, epoch=3, round=4)
        assertAlmostEqual(tl.as_float(), 1.2)

        # test random
        tl = Timeline(epoch=5, round=2, rounds_per_epoch=3)
        assertAlmostEqual(tl.as_float(), 4.66, delta=0.01)

    def test_string(self):
        tl = Timeline(epoch=2, round=2, rounds_per_epoch=10)
        assertEqual(f"{tl}", "(epoch = 2, round = 2, global round = 12)")

        tl = Timeline(global_round=12, rounds_per_epoch=10)
        assertEqual(f"{tl}", "(epoch = 2, round = 2, global round = 12)")

        tl = Timeline(global_round=10, rounds_per_epoch=10)
        assertEqual(f"{tl}", "(epoch = 1, round = 10, global round = 10)")

    def test_tick_simple(self):
        # test every epoch, i.e.  tick_interval = 1
        tl = Timeline(epoch=1, round=1, rounds_per_epoch=10)
        assertFalse(tl.tick(tick_interval=1))

        tl = Timeline(epoch=1, round=10, rounds_per_epoch=10)
        assertTrue(tl.tick(tick_interval=1))

        tl = Timeline(epoch=2, round=1, rounds_per_epoch=10)
        assertFalse(tl.tick(tick_interval=1))

        tl = Timeline(epoch=2, round=10, rounds_per_epoch=10)
        assertTrue(tl.tick(tick_interval=1))

        # test every 2 epoch
        tl = Timeline(epoch=1, round=1, rounds_per_epoch=10)
        assertFalse(tl.tick(tick_interval=2))

        tl = Timeline(epoch=1, round=10, rounds_per_epoch=10)
        assertFalse(tl.tick(tick_interval=2))

        tl = Timeline(epoch=2, round=1, rounds_per_epoch=10)
        assertFalse(tl.tick(tick_interval=2))

        tl = Timeline(epoch=2, round=10, rounds_per_epoch=10)
        assertTrue(tl.tick(tick_interval=2))

        # test every 0.2 epoch
        tl = Timeline(epoch=1, round=1, rounds_per_epoch=10)
        assertFalse(tl.tick(tick_interval=0.2))

        tl = Timeline(epoch=1, round=2, rounds_per_epoch=10)
        assertTrue(tl.tick(tick_interval=0.2))

        tl = Timeline(epoch=2, round=1, rounds_per_epoch=10)
        assertFalse(tl.tick(tick_interval=0.2))

        tl = Timeline(epoch=1, round=2, rounds_per_epoch=10)
        assertTrue(tl.tick(tick_interval=0.2))

        tl = Timeline(epoch=1, round=3, rounds_per_epoch=10)
        assertFalse(tl.tick(tick_interval=0.2))

        tl = Timeline(epoch=1, round=4, rounds_per_epoch=10)
        assertTrue(tl.tick(tick_interval=0.2))

        # test with global round
        tl = Timeline(global_round=4, rounds_per_epoch=2)
        assertFalse(tl.tick(tick_interval=3))
        assertTrue(tl.tick(tick_interval=2))

    def test_tick_complex(self):
        # doing 100 epochs, 10 rounds per epoch, ticking interval = 10 epochs
        sum = 0
        for e in range(1, 101):
            for r in range(1, 11):
                tl = Timeline(epoch=e, round=r, rounds_per_epoch=10)
                sum += tl.tick(10)
        assertEqual(sum, 10)

        # doing 100 epochs, 10 rounds per epoch, ticking interval = 0.9 epochs
        sum = 0
        for e in range(1, 101):
            for r in range(1, 11):
                tl = Timeline(epoch=e, round=r, rounds_per_epoch=10)
                sum += tl.tick(0.9)
        assertEqual(sum, 111)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import pytest
import torch
from flsim.common.pytest_helper import (
    assertAlmostEqual,
    assertTrue,
    assertEqual,
    assertFalse,
)
from flsim.privacy.user_update_clip import UserUpdateClipper
from flsim.tests import utils


@pytest.fixture(scope="class")
def prepare_user_update_clipper_test(request):
    request.cls.user_update_clipper = UserUpdateClipper()

    def _init_user_model(param_value):
        user_model = utils.TwoFC()
        user_model.fill_all(param_value)  # fill all parameters with a number
        return user_model

    request.cls.previous_user_model = _init_user_model(param_value=6.0)
    request.cls.updated_user_model = _init_user_model(param_value=7.0)
    request.cls.original_model_diff = request.cls.user_update_clipper.calc_model_diff(
        request.cls.updated_user_model, request.cls.previous_user_model
    )
    request.cls.original_model_diff_params = [
        p for p in request.cls.original_model_diff.parameters() if p.requires_grad
    ]


@pytest.mark.usefixtures("prepare_user_update_clipper_test")
class TestUserUpdateClipper:
    def _init_clipped_model_diff(self, max_norm):
        clipped_model_diff = utils.TwoFC()
        clipped_model_diff.load_state_dict(self.original_model_diff.state_dict())

        self.user_update_clipper.clip(clipped_model_diff, max_norm)
        return clipped_model_diff

    def test_calc_clip_factor(self):
        """
        Tests that the clip factor for user updates is calculated correctly.
        """
        clip_factor = self.user_update_clipper._calc_clip_factor(
            max_norm=5, per_user_norm=10
        )
        assertAlmostEqual(clip_factor, 0.5, places=3)

        clip_factor = self.user_update_clipper._calc_clip_factor(
            max_norm=1, per_user_norm=1
        )
        assertAlmostEqual(clip_factor, 1, places=3)

        clip_factor = self.user_update_clipper._calc_clip_factor(
            max_norm=2, per_user_norm=0.8
        )
        assertAlmostEqual(clip_factor, 1, places=3)

    def test_calc_user_update_norm(self):
        """
        Tests that the user update l-2 norms are calculated correctly.
        """
        model = utils.TwoFC()
        model.fill_all(2.0)
        model_params = model.parameters()
        # norm = sqrt(21*2^2)=sqrt(84)=9.16515138991168
        norm = self.user_update_clipper._calc_norm(model_params)
        assertTrue(torch.allclose(norm, torch.tensor(9.16515138991168), rtol=1e-06))

        model.fill_all(1.0)
        model_params = model.parameters()
        # norm = sqrt(21*1^2)=sqrt(21)=4.58257569495584
        norm = self.user_update_clipper._calc_norm(model_params)
        assertTrue(torch.allclose(norm, torch.tensor(4.58257569495584), rtol=1e-06))

    def test_clipped_updates_are_smaller(self):
        """
        Tests that user updates are clipped and their value is smaller than
        the original updates
        """

        # assert the parameters of model_diff are all = (7 - 6 = 1)
        for p in self.original_model_diff_params:
            assertTrue(torch.allclose(p.float(), torch.tensor(1.0)))

        clipped_model_diff = self._init_clipped_model_diff(0.0003)
        clipped_model_diff_params = [
            p for p in clipped_model_diff.parameters() if p.requires_grad
        ]
        for original, clipped in zip(
            self.original_model_diff_params, clipped_model_diff_params
        ):
            assertTrue(torch.all(original.gt(clipped)))

    def test_clipped_user_updates_non_zero(self):
        """
        Tests that user updates are not zero by clipping
        """
        clipped_model_diff = self._init_clipped_model_diff(0.0003)
        clipped_model_diff_params = [
            p for p in clipped_model_diff.parameters() if p.requires_grad
        ]
        for clipped in clipped_model_diff_params:
            allzeros = torch.zeros_like(clipped)
            assertFalse(torch.allclose(clipped, allzeros))

    def test_clipping_to_high_value_does_not_clip(self):
        """
        Tests that when clip value is set too high, user
        updates are not clipped
        """
        clipped_model_diff = self._init_clipped_model_diff(9999)
        mismatched = utils.verify_models_equivalent_after_training(
            self.original_model_diff, clipped_model_diff
        )
        assertEqual(mismatched, "")



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from hydra.core.config_store import ConfigStore

from .data_sharder import (
    RandomSharderConfig,
    SequentialSharderConfig,
    RoundRobinSharderConfig,
    BroadcastSharderConfig,
    ColumnSharderConfig,
    PowerLawSharderConfig,
)

ConfigStore.instance().store(
    name="base_random_sharder",
    node=RandomSharderConfig,
    group="sharder",
)

ConfigStore.instance().store(
    name="base_sequential_sharder",
    node=SequentialSharderConfig,
    group="sharder",
)

ConfigStore.instance().store(
    name="base_round_robin_sharder",
    node=RoundRobinSharderConfig,
    group="sharder",
)

ConfigStore.instance().store(
    name="base_column_sharder",
    node=ColumnSharderConfig,
    group="sharder",
)

ConfigStore.instance().store(
    name="base_broadcast_sharder",
    node=BroadcastSharderConfig,
    group="sharder",
)

ConfigStore.instance().store(
    name="base_power_law_sharder",
    node=PowerLawSharderConfig,
    group="sharder",
)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import math

import pytest
from flsim.common.pytest_helper import assertEqual
from flsim.optimizers.sync_aggregators import (
    FedAdamSyncAggregator,
    FedLAMBSyncAggregator,
    FedAdamSyncAggregatorConfig,
    FedLAMBSyncAggregatorConfig,
)
from flsim.tests.utils import MockQuadratic1DFL, Quadratic1D
from omegaconf import OmegaConf


def adjust_learning_rate(optimizer, new_lr):
    for param_group in optimizer.param_groups:
        param_group["lr"] = new_lr


def _test_lamb_multiple_steps(test_case, weight_decay=0):
    """
        a toy optimization example:
            min f(x) = 100 x^2 - 1

    minima is x=0.0, x is initialized at 1.0.
    """
    # set up quadratic parabola objective and optimizer
    beta1 = 0.9
    beta2 = 0.999
    eps = 1e-8

    dict_config_lamb = {
        "lr": 0.01,
        "beta1": beta1,
        "beta2": beta2,
        "eps": eps,
        "weight_decay": weight_decay,
    }
    lamb_aggregator = FedLAMBSyncAggregator(
        **OmegaConf.structured(FedLAMBSyncAggregatorConfig(**dict_config_lamb)),
        global_model=test_case.quadratic1D_lamb,
    )

    dict_config_adam = {"lr": 0.01, "weight_decay": weight_decay, "eps": eps}
    adam_aggregator = FedAdamSyncAggregator(
        **OmegaConf.structured(FedAdamSyncAggregatorConfig(**dict_config_adam)),
        global_model=test_case.quadratic1D_adam,
    )

    m_t = 0.0
    v_t = 0.0

    for i in range(1, 11):
        lamb_aggregator.optimizer.zero_grad()
        metrics = test_case.quadratic1D_lamb.fl_forward()
        loss_lamb = metrics.loss
        loss_lamb.backward()

        original_param_value = (
            test_case.quadratic1D_lamb.fl_get_module().state_dict()["x"].item()
        )
        weight_norm = abs(original_param_value)
        lamb_aggregator.optimizer.step()
        updated_param_value_lamb = (
            test_case.quadratic1D_lamb.fl_get_module().state_dict()["x"].item()
        )

        g_t = list(test_case.quadratic1D_lamb.fl_get_module().parameters())[
            0
        ].grad.data.item()

        bias_correction1 = 1.0 - beta1 ** i
        bias_correction2 = 1.0 - beta2 ** i

        m_t = beta1 * m_t + (1.0 - beta1) * g_t
        v_t = beta2 * v_t + (1.0 - beta2) * g_t ** 2

        lamb_update_unnormalized = (m_t / bias_correction1) / (
            math.sqrt(v_t / bias_correction2) + eps
        )
        lamb_update_norm = abs(lamb_update_unnormalized)

        equivalent_adam_lr = dict_config_lamb["lr"] * weight_norm / lamb_update_norm

        # we can't initialize a new adam each time because it would lose its momentum
        adjust_learning_rate(adam_aggregator.optimizer, equivalent_adam_lr)

        adam_aggregator.optimizer.zero_grad()
        metrics = test_case.quadratic1D_adam.fl_forward()
        loss_adam = metrics.loss
        loss_adam.backward()
        original_param_value = (
            test_case.quadratic1D_adam.fl_get_module().state_dict()["x"].item()
        )
        adam_aggregator.optimizer.step()
        updated_param_value_adam = (
            test_case.quadratic1D_adam.fl_get_module().state_dict()["x"].item()
        )

        assertEqual(updated_param_value_lamb, updated_param_value_adam)


@pytest.fixture(scope="class")
def prepare_lamb_optimizer_test(request):
    request.cls.quadratic1D_lamb = MockQuadratic1DFL(Quadratic1D())
    request.cls.quadratic1D_adam = MockQuadratic1DFL(Quadratic1D())


@pytest.mark.usefixtures("prepare_lamb_optimizer_test")
class TestLambOptimizer:
    def test_lamb_no_weight_decay(self):
        _test_lamb_multiple_steps(self, weight_decay=0.0)

    def test_lamb_weight_decay(self):
        _test_lamb_multiple_steps(self, weight_decay=0.1)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from hydra.core.config_store import ConfigStore  #  @manual

from .diverse_user_selector import (
    DiversityReportingUserSelectorConfig,
    DiversityStatisticsReportingUserSelectorConfig,
    DiversityMaximizingUserSelectorConfig,
)
from .simple_user_selector import (
    UniformlyRandomActiveUserSelectorConfig,
    SequentialActiveUserSelectorConfig,
    RandomRoundRobinActiveUserSelectorConfig,
    NumberOfSamplesActiveUserSelectorConfig,
    HighLossActiveUserSelectorConfig,
)


ConfigStore.instance().store(
    name="base_uniformly_random_active_user_selector",
    node=UniformlyRandomActiveUserSelectorConfig,
    group="active_user_selector",
)


ConfigStore.instance().store(
    name="base_sequential_active_user_selector",
    node=SequentialActiveUserSelectorConfig,
    group="active_user_selector",
)


ConfigStore.instance().store(
    name="base_random_round_robin_active_user_selector",
    node=RandomRoundRobinActiveUserSelectorConfig,
    group="active_user_selector",
)


ConfigStore.instance().store(
    name="base_number_of_samples_active_user_selector",
    node=NumberOfSamplesActiveUserSelectorConfig,
    group="active_user_selector",
)

ConfigStore.instance().store(
    name="base_high_loss_active_user_selector",
    node=HighLossActiveUserSelectorConfig,
    group="active_user_selector",
)

ConfigStore.instance().store(
    name="base_diversity_reporting_user_selector",
    node=DiversityReportingUserSelectorConfig,
    group="active_user_selector",
)

ConfigStore.instance().store(
    name="base_diversity_statistics_reporting_user_selector",
    node=DiversityStatisticsReportingUserSelectorConfig,
    group="active_user_selector",
)

ConfigStore.instance().store(
    name="base_diversity_maximizing_user_selector",
    node=DiversityMaximizingUserSelectorConfig,
    group="active_user_selector",
)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import logging
from dataclasses import dataclass

import torch
from flsim.channels.base_channel import IdentityChannel, FLChannelConfig
from flsim.channels.message import Message
from flsim.common.logger import Logger
from flsim.interfaces.model import IFLModel
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.count_sketch import CountSketch


class SketchChannel(IdentityChannel):
    """
    Implements a channel which sketches/compresses the model sent
    from client to server.
    """

    logger: logging.Logger = Logger.get_logger(__name__)
    random_seed = None

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=SketchChannelConfig,
            **kwargs,
        )
        super().__init__(**kwargs)
        self.num_col = self.cfg.num_col
        self.num_hash = self.cfg.num_hash
        self.prime = self.cfg.prime
        self.independence = self.cfg.independence
        self.h = torch.randint(
            low=1, high=self.cfg.prime, size=(self.cfg.num_hash, self.cfg.independence)
        )
        self.g = torch.randint(
            low=1, high=self.cfg.prime, size=(self.cfg.num_hash, self.cfg.independence)
        )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _calc_message_size_client_to_server(self, message: Message):
        assert (
            message.count_sketch is not None
        ), "Count sketch is none. Please init count_sketch in Message."
        return message.count_sketch.get_size_in_bytes()

    def _on_client_before_transmission(self, message: Message) -> Message:
        """
        Curerntly sends whole CountSketch class. We could only share relevant attributes
        to re-instantiate the class on the server side, TODO in a next refactor.
        """
        message.populate_state_dict()
        device = next(iter(message.model_state_dict.values())).device
        cs = CountSketch(
            width=self.num_col,
            depth=self.num_hash,
            prime=self.prime,
            independence=self.independence,
            h=self.h,
            g=self.g,
            device=device,
        )
        cs.sketch_state_dict(message.model_state_dict)
        message.count_sketch = cs

        return message

    def _during_transmission_client_to_server(self, message: Message) -> Message:
        if self.stats_collector:
            message_size_bytes = self._calc_message_size_client_to_server(message)
            self.stats_collector.collect_channel_stats(
                message_size_bytes, client_to_server=True
            )
        return message


@dataclass
class SketchChannelConfig(FLChannelConfig):
    _target_: str = fullclassname(SketchChannel)
    # number of columns in the table
    num_col: int = 10000
    # number of hash functions to use, also the number of rows in the table
    num_hash: int = 11
    # prime number used in hash functions
    prime: int = 2 ** 31 - 1
    independence: int = 2



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import logging
import sys


LOGGING_LEVEL = logging.WARNING
# If this flag is true logging will be printed to both stdout and stderr
PRINT_TO_STDOUT = False


class Logger:
    parent_name = "FLSimLogger"
    _instance = None
    logging_level = LOGGING_LEVEL
    print_to_stdout = PRINT_TO_STDOUT
    parent_logger = logging.getLogger(parent_name)
    parent_logger.setLevel(logging_level)
    children_loggers = []

    @classmethod
    def get_logger(cls, name: str) -> logging.Logger:
        """Returns a Logger which is a child of the PARENT_LOGGER. Logger hierarchy is
        through the name, with `.` used to denote hierarchy levels.
        """
        logger = logging.getLogger(cls.parent_name + "." + name)
        if cls.print_to_stdout:
            handler = logging.StreamHandler(sys.stdout)
            cls.parent_logger.addHandler(handler)

        cls.children_loggers.append(logger)
        return logger

    @classmethod
    def set_logging_level(cls, level: int) -> None:
        cls.parent_logger.setLevel(level)
        for child_logger in cls.children_loggers:
            child_logger.setLevel(0)
